
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, IO, Mapping
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_SYNC import Submission
    from ...models.comment_SYNC import Comment
    from ...dto_models.submission import GalleryItem

from functools import cached_property

from ...model_loaders.submission_SYNC import load_submission
from ...models.media_upload_lease import MediaUploadLease
from ...model_loaders.media_upload_lease import load_media_upload_lease
from ...http.payload import guess_mimetype_from_filename
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_iterator import CallChunkCallingIterator
from ...iterators.call_chunk_chaining_iterator import CallChunkChainingIterator
from ...iterators.call_chunk import CallChunk
from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.submission_sync1 import SearchSubmissionsListingPaginator, SubmissionDuplicatesPaginator
from ...model_loaders.comment_SYNC import load_comment
from .fetch_SYNC import Fetch
from .get_SYNC import Get

class SubmissionProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.fetch: Fetch = Fetch(self, client)
        self.get: Get = Get(client)

    def bulk_fetch(self, ids: Iterable[int]) -> CallChunkChainingIterator[Submission]:
        def mass_fetch(ids: Sequence[int]) -> Sequence[Submission]:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_submission(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingIterator(CallChunk(mass_fetch, chunk) for chunk in chunked(ids, 100))

    def reply(self, submission_id: int, text: str) -> Comment:
        data = {
            'thing_id': 't3_' + to_base36(submission_id),
            'text': text,
            'return_rtjson': '1',
        }
        result = self._client.request('POST', '/api/comment', data=data)
        return load_comment(result, self._client)

    class _upload_media:
        def __init__(self, outer: SubmissionProcedures):
            self._client = outer._client

        def __call__(self, file: IO[bytes]) -> MediaUploadLease:
            return self.upload(file)

        def obtain_upload_lease(self, *, filename: str, mimetype: Optional[str] = None) -> MediaUploadLease:
            if mimetype is None:
                mimetype = guess_mimetype_from_filename(filename)
            result = self._client.request('POST', '/api/media/asset',
                    data={'filepath': filename, 'mimetype': mimetype})
            return load_media_upload_lease(result)

        def deposit_file(self, file: IO[bytes], upload_lease: MediaUploadLease, *,
                timeout: float = 1000) -> None:
            session = self._client.http.session
            resp = session.request('POST', upload_lease.endpoint, data=upload_lease.fields,
                    files={'file': file}, timeout=timeout)
            resp.raise_for_status()

        def upload(self, file: IO[bytes], *, timeout: float = 1000) -> MediaUploadLease:
            upload_lease = self.obtain_upload_lease(filename=file.name)
            self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

    upload_media: cached_property[_upload_media] = cached_property(_upload_media)

    def create_text_post(self,
        sr: str,
        title: str,
        text: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
    ) -> int:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'self')
            yield ('sr', sr)
            yield ('title', title)
            yield ('text', text)
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if original_content: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = self._client.request('POST', '/api/submit', data=dict(g()))
        return int(root['json']['data']['id'], 36)

    def create_link_post(self,
        sr: str,
        title: str,
        url: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
        resubmit: bool = True,
    ) -> int:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'link')
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', url)
            if resubmit: yield ('resubmit', '1')
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if original_content: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = self._client.request('POST', '/api/submit', data=dict(g()))
        return int(root['json']['data']['id'], 36)

    def create_image_post(self,
        sr: str,
        title: str,
        image_url: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
    ) -> None:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'image')
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', image_url)
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if original_content: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        self._client.request('POST', '/api/submit', data=dict(g()))

    def create_video_post(self,
        sr: str,
        title: str,
        video_url: str,
        thumbnail_url: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
        vgif: bool = False,
    ) -> None:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'videogif' if vgif else 'video')
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', video_url)
            yield ('video_poster_url', thumbnail_url)
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if original_content: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        self._client.request('POST', '/api/submit', data=dict(g()))

    def create_gallery_post(self,
        sr: str,
        title: str,
        items: Sequence[GalleryItem],
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
    ) -> int:
        gallery_items_data: Sequence[Mapping[str, str]] = [
            {
                'media_id': m.media_id,
                'caption': m.caption,
                'outbound_url': m.outbound_link,
            }
            for m in items
        ]

        def g() -> Iterable[tuple[str, object]]:
            yield ('sr', sr)
            yield ('title', title)
            yield ('items', gallery_items_data)
            yield ('sendreplies', reply_notifications)
            if spoiler: yield ('spoiler', True)
            if nsfw: yield ('nsfw', True)
            if original_content: yield ('original_content', True)
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = self._client.request('POST', '/api/submit_gallery_post', json=dict(g()))
        return int(root['json']['data']['id'][3:], 36)

    def create_poll_post(self,
        sr: str,
        title: str,
        text: str,
        options: Sequence[str],
        duration: int,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
    ) -> int:
        def g() -> Iterable[tuple[str, object]]:
            yield ('sr', sr)
            yield ('title', title)
            yield ('text', text)
            yield ('options', options)
            yield ('duration', duration)
            yield ('sendreplies', reply_notifications)
            if spoiler: yield ('spoiler', True)
            if nsfw: yield ('nsfw', True)
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = self._client.request('POST', '/api/submit_poll_post', json=dict(g()))
        return int(root['json']['data']['id'][3:], 36)

    def crosspost(self,
        sr: str,
        title: str,
        submission_id: int,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        original_content: bool = False,
        collection_uuid: str = '',
        flair_uuid: str = '',
        flair_text: str = '',
        event_start: str = '',
        event_end: str = '',
        event_tz: str = '',
    ) -> int:
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'self')
            yield ('sr', sr)
            yield ('title', title)
            yield ('crosspost_parent', 't3_' + to_base36(submission_id))
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if original_content: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = self._client.request('POST', '/api/submit', data=dict(g()))
        return int(root['json']['data']['id'], 36)

    def edit_post_text(self, submission_id: int, text: str) -> Submission:
        data = {
            'thing_id': 't3_' + to_base36(submission_id),
            'text': text,
            'return_rtjson': '1',
        }
        result = self._client.request('POST', '/api/editusertext', data=data)
        return load_submission(result, self._client)

    def delete(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/del', data=data)

    def lock(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/lock', data=data)

    def unlock(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/unlock', data=data)

    def vote(self, submission_id: int, direction: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'dir': str(direction),
        }
        self._client.request('POST', '/api/vote', data=data)

    def save(self, submission_id: int, category: Optional[str] = None) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
        }
        if category is not None:
            data['category'] = category
        self._client.request('POST', '/api/save', data=data)

    def unsave(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/unsave', data=data)

    def hide(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/hide', data=data)

    def unhide(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/unhide', data=data)

    def bulk_hide(self, submission_ids: Iterable[int]) -> CallChunkCallingIterator[None]:
        def mass_hide(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            self._client.request('POST', '/api/hide', data={'id': ids_str})

        return CallChunkCallingIterator(CallChunk(mass_hide, chunk) for chunk in chunked(submission_ids, 300))

    def bulk_unhide(self, submission_ids: Iterable[int]) -> CallChunkCallingIterator[None]:
        def mass_unhide(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            self._client.request('POST', '/api/unhide', data={'id': ids_str})

        return CallChunkCallingIterator(CallChunk(mass_unhide, chunk) for chunk in chunked(submission_ids, 300))

    def mark_nsfw(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/marknsfw', data=data)

    def unmark_nsfw(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/unmarknsfw', data=data)

    def mark_spoiler(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/spoiler', data=data)

    def unmark_spoiler(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/unspoiler', data=data)

    def distinguish(self, submission_id: int) -> Submission:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'how': 'yes',
        }
        root = self._client.request('POST', '/api/distinguish', data=data)
        return load_submission(root['json']['data']['things'][0]['data'], self._client)

    def undistinguish(self, submission_id: int) -> Submission:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'how': 'no',
        }
        root = self._client.request('POST', '/api/distinguish', data=data)
        return load_submission(root['json']['data']['things'][0]['data'], self._client)

    def sticky(self, submission_id: int, slot: Optional[int] = None) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'state': '1',
        }
        if slot is not None:
            data['num'] = str(slot)
        self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    def unsticky(self, submission_id: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'state': '0',
        }
        self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    def pin_to_profile(self, submission_id: int, slot: Optional[int] = None) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'to_profile': '1',
            'state': '1',
        }
        if slot is not None:
            data['num'] = str(slot)
        self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    def unpin_from_profile(self, submission_id: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'to_profile': '1',
            'state': '0',
        }
        self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    def set_contest_mode(self, submission_id: int, state: bool) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'state': '01'[state],
        }
        self._client.request('POST', '/api/set_contest_mode', data=data)

    def set_suggested_sort(self, submission_id: int, sort: Optional[str]) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
        }
        if sort is not None:
            data['sort'] = sort
        self._client.request('POST', '/api/set_suggested_sort', data=data)

    def enable_reply_notifications(self, submission_id: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'state': '1'
        }
        self._client.request('POST', '/api/sendreplies', data=data)

    def disable_reply_notifications(self, submission_id: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'state': '0'
        }
        self._client.request('POST', '/api/sendreplies', data=data)

    def set_event_time(self, submission_id: int,
            event_start: str, event_end: str, event_tz: str) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'event_start': event_start,
            'event_end': event_end,
            'event_tz': event_tz,
        }
        self._client.request('POST', '/api/event_post_time', data=data)

    def follow_event(self, submission_id: int) -> None:
        data = {
            'fullname': 't3_' + to_base36(submission_id),
            'follow': '1',
        }
        self._client.request('POST', '/api/follow_post', data=data)

    def unfollow_event(self, submission_id: int) -> None:
        data = {
            'fullname': 't3_' + to_base36(submission_id),
            'follow': '0',
        }
        self._client.request('POST', '/api/follow_post', data=data)

    def approve(self, submission_id: int) -> None:
        data = {'id': 't3_' + to_base36(submission_id)}
        self._client.request('POST', '/api/approve', data=data)

    def remove(self, submission_id: int) -> None:
        data = {
            'id': 't3_' + to_base36(submission_id),
            'spam': '0',
        }
        self._client.request('POST', '/api/remove', data=data)

    def ignore_reports(self, idn: int) -> None:
        self._client.request('POST', '/api/ignore_reports', data={'id': 't3_' + to_base36(idn)})

    def unignore_reports(self, idn: int) -> None:
        self._client.request('POST', '/api/unignore_reports', data={'id': 't3_' + to_base36(idn)})

    def snooze_reports(self, idn: int, reason: str) -> None:
        data = {'id': 't3_' + to_base36(idn), 'reason': reason}
        self._client.request('POST', '/api/snooze_reports', data=data)

    def unsnooze_reports(self, idn: int, reason: str) -> None:
        data = {'id': 't3_' + to_base36(idn), 'reason': reason}
        self._client.request('POST', '/api/unsnooze_reports', data=data)

    def apply_removal_reason(self,
            submission_id: int,
            reason_id: Optional[str],
            note: Optional[str] = None) -> None:
        target = 't3_' + to_base36(submission_id)
        json_data = {'item_ids': [target], 'reason_id': reason_id, 'mod_note': note}
        self._client.request('POST', '/api/v1/modactions/removal_reasons', json=json_data)

    def send_removal_comment(self,
            submission_id: int,
            title: str,
            message: str) -> Comment:
        target = 't3_' + to_base36(submission_id)
        json_data = {
            'type': 'public',
            'item_id': [target],
            'title': title,
            'message': message,
        }
        root = self._client.request('POST', '/api/v1/modactions/removal_link_message', json=json_data)
        return load_comment(root, self._client)

    def send_removal_message(self,
            submission_id: int,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        target = 't3_' + to_base36(submission_id)
        json_data = {
            'type': 'private' + ('_exposed' if exposed else ''),
            'item_id': [target],
            'title': title,
            'message': message,
        }
        self._client.request('POST', '/api/v1/modactions/removal_link_message', json=json_data)

    def search(self, sr: str, query: str, amount: Optional[int] = None, *,
        sort: str = 'relevance', time: str = 'all',
    ) -> ImpartedPaginatorChainingIterator[SearchSubmissionsListingPaginator, Submission]:
        uri = '/search'
        if sr:
            uri = f'/r/{sr}/search'
        p = SearchSubmissionsListingPaginator(
                self._client, uri,
                params={'q': query, 'restrict_sr': '1'},
                    sort=sort, time=time)
        return ImpartedPaginatorChainingIterator(p, amount)

    def duplicates(self, target: int, amount: Optional[int] = None, *,
        sort: str = 'num_comments',
    ) -> ImpartedPaginatorChainingIterator[SubmissionDuplicatesPaginator, Submission]:
        subm_id = to_base36(target)
        p = SubmissionDuplicatesPaginator(self._client, f'/duplicates/{subm_id}', sort=sort)
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, MutableMapping
if TYPE_CHECKING:
    from ..http.session_base_SYNC import SessionBase
    from ..http.requestor_SYNC import Requestor

from ..auth.const import RESOURCE_BASE_URL, TOKEN_OBTAINMENT_URL
from ..http.http_client_SYNC import BasicRequestDefaultsHTTPClient
from ..auth.typedefs import AuthorizationGrant
from .authorizer_SYNC import Authorized
from .rate_limited_SYNC import RateLimited
from .recorded_SYNC import Recorded
from .reddit_give_me_json_please_SYNC import RedditGiveMeJSONPlease
from .reddit_token_obtainment_client_SYNC import RedditTokenObtainmentClient
from ..http.transport.SYNC import new_session
from ..http.util.case_insensitive_dict import CaseInsensitiveDict
from ..util.user_agent_SYNC import get_user_agent_from_session
from ..auth.token import Token
from .recorded_SYNC import Last
from .authorizer_SYNC import Authorizer


class RedditHTTPClient(BasicRequestDefaultsHTTPClient):
    @property
    def user_agent(self) -> str:
        return self.headers.get('User-Agent', '')

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self.headers['User-Agent'] = value

    @user_agent.deleter
    def user_agent(self) -> None:
        self.headers.pop('User-Agent', None)

    @property
    def last(self) -> Last:
        return self.get_last()

    def __init__(self,
        session: SessionBase,
        requestor: Optional[Requestor] = None,
        *,
        params: Optional[MutableMapping[str, str]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
        last: Optional[Last] = None,
    ) -> None:
        super().__init__(session, requestor, params=params, headers=headers)
        self.base_url: str = RESOURCE_BASE_URL
        self.user_agent_lead: str = ''
        self._last: Optional[Last] = last

    def get_last(self) -> Last:
        if self._last is None:
            raise RuntimeError('value not set')
        return self._last

    def set_last(self, value: Last) -> None:
        self._last = value


class PublicAPIRedditHTTPClient(RedditHTTPClient):
    @property
    def authorizer(self) -> Authorizer:
        return self.get_authorizer()

    def __init__(self,
        session: SessionBase,
        requestor: Optional[Requestor] = None,
        *,
        params: Optional[MutableMapping[str, str]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
        last: Optional[Last] = None,
        authorizer: Optional[Authorizer] = None,
    ) -> None:
        super().__init__(session, requestor, params=params, headers=headers, last=last)
        self._authorizer: Optional[Authorizer] = authorizer

    def get_authorizer(self) -> Authorizer:
        if self._authorizer is None:
            raise RuntimeError('value not set')
        return self._authorizer

    def set_authorizer(self, value: Authorizer) -> None:
        self._authorizer = value


def build_public_api_reddit_http_client(
    client_id: str,
    client_secret: str,
    grant: AuthorizationGrant,
    *,
    session: Optional[SessionBase] = None,
) -> PublicAPIRedditHTTPClient:
    if session is None:
        session = new_session()
    ua = get_user_agent_from_session(session)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    recorder = Recorded(session)
    last = Last(recorder)
    token_client = RedditTokenObtainmentClient(
        session,
        TOKEN_OBTAINMENT_URL,
        (client_id, client_secret),
        grant,
        headers=headers,
    )
    authorizer = Authorizer(token_client)
    requestor = RedditGiveMeJSONPlease(RateLimited(Authorized(recorder, authorizer)))
    http = PublicAPIRedditHTTPClient(session, requestor, headers=headers, authorizer=authorizer, last=last)
    http.user_agent_lead = ua
    return http

def build_public_api_reddit_http_client_from_access_token(
    access_token: str,
    *,
    session: Optional[SessionBase] = None,
) -> PublicAPIRedditHTTPClient:
    if session is None:
        session = new_session()
    ua = get_user_agent_from_session(session)
    headers = CaseInsensitiveDict({'User-Agent': ua})
    recorder = Recorded(session)
    last = Last(recorder)
    authorizer = Authorizer(token=Token(access_token))
    requestor = RedditGiveMeJSONPlease(RateLimited(Authorized(recorder, authorizer)))
    http = PublicAPIRedditHTTPClient(session, requestor, headers=headers, authorizer=authorizer, last=last)
    return http




from __future__ import annotations
from typing import TypeVar, Iterator, Generic, Optional, Sequence

from .paginator import Paginator

E = TypeVar('E')

class PaginatorChainingIterator(Iterator[E]):
    def __init__(self, paginator: Paginator[E], amount: Optional[int] = None) -> None:
        self._paginator: Paginator[E] = paginator
        self.remaining: Optional[int] = amount
        self.current_iter: Iterator[E] = iter(())
        self._pagination_iterator: Iterator[Sequence[E]] = iter(paginator)

    def __iter__(self) -> Iterator[E]:
        return self

    def __next__(self) -> E:
        limit = self._paginator.limit
        remaining = self.remaining
        if remaining is None or remaining > 0:
            while True:
                for elem in self.current_iter:
                    if self.remaining is not None:
                        self.remaining -= 1
                    return elem

                if (limit is not None and remaining is not None) and limit > remaining:
                    self._paginator.limit = remaining

                it = next(self._pagination_iterator)
                self.current_iter = iter(it)

        raise StopIteration



__bound = 'Paginator[E]'
TPaginator = TypeVar('TPaginator', bound=Paginator)  # type: ignore[type-arg]

class ImpartedPaginatorChainingIterator(PaginatorChainingIterator[E], Generic[TPaginator, E]):
    def __init__(self, paginator: TPaginator, amount: Optional[int] = None) -> None:
        super().__init__(paginator, amount)
        self.__paginator: TPaginator = paginator

    def get_paginator(self) -> TPaginator:
        return self.__paginator




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Iterable
if TYPE_CHECKING:
    from ...request import Request
    from ...response import Response

import asyncio

import aiohttp  # type: ignore[import]

from ..ASYNC import register
from ...session_base_ASYNC import SessionBase
from ... import exceptions
from ... import payload
from ...response import UResponse
from ...util.case_insensitive_dict import CaseInsensitiveDict

def _generate_request_kwargs(r: Request, etv: float) -> Iterable[tuple[str, Any]]:
    client_timeout = aiohttp.ClientTimeout(
        total=etv,
        connect=etv,
        sock_connect=etv,
        sock_read=etv,
    )
    if etv == -1:
        client_timeout = aiohttp.ClientTimeout(
            total=None,
            connect=None,
            sock_connect=None,
            sock_read=None,
        )
    yield ('timeout', client_timeout)

    yield ('method', r.verb)
    yield ('url', r.uri)
    yield ('params', r.params)

    headers = dict(r.headers)
    pld = r.payload
    if pld is None:
        pass

    elif isinstance(pld, payload.Bytes):
        pld.apply_content_type(headers)
        yield ('data', pld.data)

    elif isinstance(pld, payload.Text):
        pld.apply_content_type(headers)
        yield ('data', pld.text.encode())

    elif isinstance(pld, payload.JSON):
        yield ('json', pld.json)

    elif isinstance(pld, payload.URLEncodedFormData):
        yield ('data', dict(pld.data))

    elif isinstance(pld, payload.MultipartFormData):
        text_plds: list[payload.MultipartTextField] = []
        file_plds: list[payload.MultipartFileField] = []
        for part in pld.parts:
            if isinstance(part, payload.MultipartTextField):
                text_plds.append(part)
            elif isinstance(part, payload.MultipartFileField):
                file_plds.append(part)

        if not file_plds:
            # aiohttp won't send a multipart if no files
            raise Exception('multipart without file fields not supported')

        formdata = aiohttp.FormData()
        for ty in text_plds:
            formdata.add_field(ty.name, ty.value)
        for fy in file_plds:
            formdata.add_field(fy.name, fy.file, filename=fy.filename, content_type=fy.content_type)

        yield ('data', formdata)

    else:
        raise Exception('unsupported payload type')

    yield ('headers', headers)


class Session(SessionBase):
    def __init__(self,
        aiohttp_client: aiohttp.ClientSession,
    ) -> None:
        super().__init__()
        self.session: aiohttp.ClientSession = aiohttp_client

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        etv = self._get_effective_timeout_value(timeout)
        kwargs = dict(_generate_request_kwargs(request, etv))
        try:
            async with self.session.request(**kwargs) as response:
                content = await response.content.read()
        except asyncio.TimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        aiohttp_headers: Any = response.headers
        headers = CaseInsensitiveDict(dict(aiohttp_headers))
        return UResponse(
            status=response.status,
            headers=headers,
            data=content,
            underlying_object=response,
        )

    async def close(self) -> None:
        await self.session.close()

def new_session() -> Session:
    connector = aiohttp.TCPConnector(limit=20)
    se = aiohttp.ClientSession(connector=connector)
    return Session(se)

name: str = aiohttp.__name__
version: str = aiohttp.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_session=new_session,
)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..models.live_thread_ASYNC import LiveThread, LiveUpdate

def load_live_thread(d: Mapping[str, Any], client: Client) -> LiveThread:
    return LiveThread(d, client)

def load_live_update(d: Mapping[str, Any], client: Client) -> LiveUpdate:
    return LiveUpdate(d, client)




from __future__ import annotations

from .listing.mixins.subreddit_detail_ASYNC import SubredditDetail
from .listing.subreddit_listing_async_paginator import SubredditListingAsyncPaginator
from ...models.subreddit_ASYNC import Subreddit

class SubscribedListingAsyncPaginator(
    SubredditDetail[Subreddit],
    SubredditListingAsyncPaginator,
): pass

class ContributingListingAsyncPaginator(
    SubredditDetail[Subreddit],
    SubredditListingAsyncPaginator,
): pass

class ModeratingListingAsyncPaginator(
    SubredditDetail[Subreddit],
    SubredditListingAsyncPaginator,
): pass




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from .live_thread_base import BaseLiveThread, BaseLiveUpdate

class LiveThread(BaseLiveThread):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

    async def configure(self, title: str, description: str, resources: str, nsfw: bool) -> None:
        await self.client.p.live_thread.configure(self.idt, title, description, resources, nsfw)

    async def get_live_update(self, update_uuid: str) -> LiveUpdate:
        return await self.client.p.live_thread.get_live_update(self.idt, update_uuid)

    async def create_live_update(self, body: str) -> None:
        await self.client.p.live_thread.create_live_update(self.idt, body)

    async def strike_live_update(self, update_uuid: str) -> None:
        await self.client.p.live_thread.strike_live_update(self.idt, update_uuid)

    async def delete_live_update(self, update_uuid: str) -> None:
        await self.client.p.live_thread.delete_live_update(self.idt, update_uuid)


class LiveUpdate(BaseLiveUpdate):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client




from __future__ import annotations
from typing import Sequence

from ..listing.listing_async_paginator import ListingAsyncPaginator
from ....model_loaders.submission_ASYNC import load_submission
from ....models.submission_ASYNC import Submission

class SubmissionListingAsyncPaginator(ListingAsyncPaginator[Submission]):
    async def fetch(self) -> Sequence[Submission]:
        data = await self._fetch_data()
        return [load_submission(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import Iterable

from ..listing.mixins.sort_SYNC import Sort
from ..listing.mixins.subreddit_detail_SYNC import SubredditDetail
from ..listing.submission_listing_paginator import SubmissionListingPaginator
from ..listing.comment_listing_paginator import ExtraSubmissionFieldsCommentListingPaginator
from ..listing.submission_and_comment_listing_paginator import SubmissionAndExtraSubmissionFieldsCommentListingPaginator
from ....models.comment_SYNC import ExtraSubmissionFieldsComment
from ....models.submission_SYNC import Submission

class OverviewListingPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingPaginator,
): pass

class CommentsListingPaginator(
    Sort[ExtraSubmissionFieldsComment],
    SubredditDetail[ExtraSubmissionFieldsComment],
    ExtraSubmissionFieldsCommentListingPaginator,
): pass

class SubmittedListingPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class GildedListingPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingPaginator,
): pass

class UpvotedListingPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class DownvotedListingPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class HiddenListingPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
): pass

class SavedListingPaginator(
    Sort[object],
    SubredditDetail[object],
    SubmissionAndExtraSubmissionFieldsCommentListingPaginator,
): pass

class SavedSubmissionsListingPaginator(
    Sort[Submission],
    SubredditDetail[Submission],
    SubmissionListingPaginator,
):
    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('type', 'links')

class SavedCommentsListingPaginator(
    Sort[ExtraSubmissionFieldsComment],
    SubredditDetail[ExtraSubmissionFieldsComment],
    ExtraSubmissionFieldsCommentListingPaginator,
):
    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('type', 'comments')




from __future__ import annotations

from ..listing.mixins.subreddit_detail_SYNC import SubredditDetail
from ..listing.subreddit_listing_paginator import SubredditListingPaginator
from ....models.subreddit_SYNC import Subreddit

class UserSubredditListingPaginator(
    SubredditDetail[Subreddit],
    SubredditListingPaginator,
): pass




from __future__ import annotations
from typing import TypeVar, Callable, Generic, Awaitable

TInput = TypeVar('TInput')
TOutput = TypeVar('TOutput')

class AsyncCallChunk(Generic[TInput, TOutput]):
    def __init__(self,
        operation: Callable[[TInput], Awaitable[TOutput]],
        operand: TInput,
    ) -> None:
        self.operation: Callable[[TInput], Awaitable[TOutput]] = operation
        self.operand: TInput = operand

    async def __call__(self) -> TOutput:
        return await self.operation(self.operand)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Tuple, Mapping, Any
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.my_account_ASYNC import MyAccount
    from ...models.user_relationship_item import UserRelationshipItem, FriendRelationshipItem
    from ...models.trophy import Trophy
    from ...models.karma_breakdown_entry import KarmaBreakdownEntry

from .pull_subreddits_ASYNC import PullSubreddits
from ...model_loaders.my_account_ASYNC import load_account
from ...model_loaders.user_relationship_item import load_user_relationship_item, load_friend_relationship_item
from ...model_loaders.karma_breakdown_entry import load_karma_breakdown_entry
from ...model_loaders.trophy import load_trophy
from ... import exceptions
from ...util.base_conversion import to_base36

class AccountProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.pull_subreddits: PullSubreddits = PullSubreddits(client)

    async def fetch(self) -> MyAccount:
        root = await self._client.request('GET', '/api/v1/me')
        if len(root) < 3:
            raise RuntimeError('no user context')
        return load_account(root, self._client)

    async def get_preferences(self) -> Mapping[str, Any]:
        return await self._client.request('GET', '/api/v1/me/prefs')

    async def set_preferences(self, json_dict: Mapping[str, Any]) -> Mapping[str, Any]:
        return await self._client.request('PATCH', '/api/v1/me/prefs', json=json_dict)

    async def get_karma_breakdown(self) -> Sequence[KarmaBreakdownEntry]:
        root = await self._client.request('GET', '/api/v1/me/karma')
        entries = root['data']
        return [load_karma_breakdown_entry(d) for d in entries]

    async def get_trophies(self) -> Sequence[Trophy]:
        root = await self._client.request('GET', '/api/v1/me/trophies')
        kind_data = root['data']['trophies']
        return [load_trophy(d['data']) for d in kind_data]

    async def get_friend(self, name: str) -> Optional[UserRelationshipItem]:
        try:
            root = await self._client.request('GET', f'/api/v1/me/friends/{name}')
        except exceptions.RedditError as e:
            if e.codename == 'USER_DOESNT_EXIST':
                return None
            raise
        return load_user_relationship_item(root)

    async def friends(self) -> Sequence[UserRelationshipItem]:
        root = await self._client.request('GET', '/api/v1/me/friends')
        entries = root['data']['children']
        return [load_user_relationship_item(d) for d in entries]

    async def add_friend(self, name: str, note: Optional[str] = None) -> FriendRelationshipItem:
        json_data = {} if note is None else {'note': note}
        root = await self._client.request('PUT', f'/api/v1/me/friends/{name}', json=json_data)
        return load_friend_relationship_item(root)

    async def remove_friend(self, name: str) -> bool:
        try:
            await self._client.request('DELETE', f'/api/v1/me/friends/{name}')
        except exceptions.RedditError as e:
            if e.codename == 'NOT_FRIEND':
                return False
            raise
        return True

    async def blocked(self) -> Sequence[UserRelationshipItem]:
        root = await self._client.request('GET', '/prefs/blocked')
        entries = root['data']['children']
        return [load_user_relationship_item(d) for d in entries]

    async def block_user_by_name(self, name: str) -> None:
        await self._client.request('POST', '/api/block_user', data={'name': name})

    async def block_user_by_id(self, idn: int) -> None:
        await self._client.request('POST', '/api/block_user', data={'account_id': to_base36(idn)})

    async def unblock_user_by_target_name(self, agent_idn: int, target_name: str) -> None:
        data = {
            'type': 'enemy',
            'container': 't2_' + to_base36(agent_idn),
            'name': target_name,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def unblock_user_by_target_id(self, agent_idn: int, target_id: int) -> None:
        data = {
            'type': 'enemy',
            'container': 't2_' + to_base36(agent_idn),
            'id': 't2_' + to_base36(target_id),
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def trusted(self) -> Sequence[UserRelationshipItem]:
        root = await self._client.request('GET', '/prefs/trusted')
        entries = root['data']['children']
        return [load_user_relationship_item(d) for d in entries]

    async def add_trusted_user(self, name: str) -> None:
        await self._client.request('POST', '/api/add_whitelisted', params={'name': name})

    async def remove_trusted_user(self, name: str) -> None:
        await self._client.request('POST', '/api/remove_whitelisted', params={'name': name})

    async def messaging(self) -> Tuple[Sequence[UserRelationshipItem], Sequence[UserRelationshipItem]]:
        root = await self._client.request('GET', '/prefs/messaging')
        blocked_entries = root[0]['data']['children']
        trusted_entries = root[1]['data']['children']
        return (
            [load_user_relationship_item(d) for d in blocked_entries],
            [load_user_relationship_item(d) for d in trusted_entries],
        )




from . import (SYNC, ASYNC)  # noqa: F401
from . import (client_SYNC, client_ASYNC)  # noqa: F401
from . import const  # noqa: F401
from . import core  # noqa: F401
from . import models  # noqa: F401
from . import paginators  # noqa: F401




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Sequence
if TYPE_CHECKING:
    from ....client_ASYNC import Client

from operator import itemgetter

from ..listing.listing_async_paginator import ListingAsyncPaginator
from ....model_loaders.user_relationship_item import load_user_relationship_item, load_banned_user_relation_item
from ....models.user_relationship_item import UserRelationshipItem, BannedUserRelationshipItem

T = TypeVar('T')

class LegacyModerationUsersAsyncPaginator(ListingAsyncPaginator[T]):
    def __init__(self,
        client: Client,
        uri: str,
        *,
        limit: Optional[int] = 100,
    ):
        super().__init__(client, uri, limit=limit, cursor_extractor=itemgetter('rel_id'))


class UserRelationshipItemListingAsyncPaginator(LegacyModerationUsersAsyncPaginator[UserRelationshipItem]):
    async def fetch(self) -> Sequence[UserRelationshipItem]:
        data = await self._fetch_data()
        return [load_user_relationship_item(d) for d in data['children']]

class BannedUserRelationshipItemListingAsyncPaginator(LegacyModerationUsersAsyncPaginator[BannedUserRelationshipItem]):
    async def fetch(self) -> Sequence[BannedUserRelationshipItem]:
        data = await self._fetch_data()
        return [load_banned_user_relation_item(d) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.comment_tree_ASYNC import SubmissionTreeNode
    from .ASYNC import CommentTreeProcedures

from ...util.base_conversion import to_base36
from ...model_loaders.comment_tree_ASYNC import load_submission_tree_node

class Fetch:
    def __init__(self, outer: CommentTreeProcedures, client: Client):
        self._outer = outer
        self._client = client

    async def __call__(self,
        submission_id: int,
        comment_id: Optional[int] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        submission_id36 = to_base36(submission_id)
        comment_id36 = comment_id if comment_id is None else to_base36(comment_id)
        return await self.by_id36(submission_id36, comment_id36, sort=sort, limit=limit, depth=depth, context=context)

    async def by_id36(self,
        submission_id36: str,
        comment_id36: Optional[str] = None,
        *,
        sort: str = 'confidence',
        limit: Optional[int] = None,
        depth: Optional[int] = None,
        context: Optional[int] = None,
    ) -> SubmissionTreeNode:
        def g() -> Iterable[tuple[str, str]]:
            if comment_id36 is not None:
                yield ('comment', comment_id36)
            if sort:
                yield ('sort', sort)
            if limit is not None:
                yield ('limit', str(limit))
            if depth is not None:
                yield ('depth', str(depth))
            if context is not None:
                yield ('context', str(context))

        root = await self._client.request('GET', '/comments/' + submission_id36, params=dict(g()))
        return load_submission_tree_node(root, self._client, sort)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from .ASYNC import SubredditProcedures

from ...util.base_conversion import to_base36
from ...exceptions import NoResultException

class Fetch:
    def __init__(self, outer: SubredditProcedures, client: Client):
        self._outer = outer
        self._client = client

    async def __call__(self, id: int) -> object:
        id36 = to_base36(id)
        return await self.by_id36(id36)

    async def by_id36(self, id36: str) -> object:
        v = await self._outer.get.by_id36(id36)
        if v is None:
            raise NoResultException('target not found')
        return v




from __future__ import annotations
from typing import Mapping, Any, Sequence, Iterator, overload, Union

from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class FlairEmojiUploadLease(IArtifact):
    d: Mapping[str, Any]
    endpoint: str
    fields: Mapping[str, str]
    s3_object_key: str
    location: str

@dataclass(repr=False, eq=False)
class FlairEmoji:
    d: Mapping[str, Any]
    name: str
    creator_id36: str
    creator_id: int
    image_url: str
    mod_only: bool
    user_enabled: bool
    post_enabled: bool


class SubredditFlairEmojis(Sequence[FlairEmoji]):
    def __init__(self,
            subreddit_emojis: Sequence[FlairEmoji],
            reddit_emojis: Sequence[FlairEmoji],
            subreddit_id36: str):
        self.subreddit_emojis: Sequence[FlairEmoji] = subreddit_emojis
        self.reddit_emojis: Sequence[FlairEmoji] = reddit_emojis
        #self.all_emojis: Sequence[FlairEmoji] = list(self.subreddit_emojis) + list(self.reddit_emojis)
        self.subreddit_id36: str = subreddit_id36
        self.subreddit_id: int = int(subreddit_id36, 36)

    def __len__(self) -> int:
        return len(self.subreddit_emojis)

    def __contains__(self, item: object) -> bool:
        return item in self.subreddit_emojis

    def __iter__(self) -> Iterator[FlairEmoji]:
        return iter(self.subreddit_emojis)

    @overload
    def __getitem__(self, index: int) -> FlairEmoji: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[FlairEmoji]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[FlairEmoji, Sequence[FlairEmoji]]:
        return self.subreddit_emojis[index]




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime, timezone

from .artifact import Artifact

class BaseSubreddit(Artifact):
    class Me:
        class MeFlair:
            def __init__(self, d: Mapping[str, Any]):
                # Value `False` if object from search item.
                self.enabled: bool = False if (x := d['user_sr_flair_enabled']) is None else x
                self.has_had_flair: bool = d['user_flair_text'] is not None
                self.bg_color: str = d['user_flair_background_color'] or ''
                user_flair_css_class_temp: Optional[str] = d['user_flair_css_class']
                self.has_had_css_class_when_no_flair_template: bool = user_flair_css_class_temp is not None
                self.css_class: str = user_flair_css_class_temp or ''
                self.position: str = d['user_flair_position']
                self.template_uuid: Optional[str] = d['user_flair_template_id']
                self.text: str = d['user_flair_text'] or ''
                self.fg_light_or_dark: str = d['user_flair_text_color'] or ''
                self.type: str = d['user_flair_type']

        def __init__(self, d: Mapping[str, Any]):
            self.favorited: bool = d['user_has_favorited']
            self.is_banned: bool = d['user_is_banned']
            self.is_contributor: bool = d['user_is_contributor']
            self.is_moderator: bool = d['user_is_moderator']
            self.is_muted: bool = d['user_is_muted']
            self.is_subscribed: bool = d['user_is_subscriber']
            self.sr_theme_enabled: bool = d['user_sr_theme_enabled']
            self.flair: BaseSubreddit.Me.MeFlair = self.MeFlair(d)

    class SubredditFlair:
        def __init__(self, d: Mapping[str, Any]):
            # Value is always `False` if object was from search.
            self.user_flairs_enabled: bool = d['user_flair_enabled_in_sr']
            self.link_flairs_enabled: bool = d['link_flair_enabled']
            self.users_can_assign_user_flair: bool = d['can_assign_user_flair']
            self.users_can_assign_link_flair: bool = d['can_assign_link_flair']
            self.user_flair_position: str = d['user_flair_position']
            self.link_flair_position: str = d['link_flair_position']

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)

        self.name: str = d['display_name']
        self.openness: str = d['subreddit_type']

        self.subscriber_count: int = d['subscribers']
        # Value `-1` if object from search item.
        self.viewing_count: int = -1 if (x := d['active_user_count']) is None else x

        self.title: str = d['title']
        self.public_description: str = d['public_description']
        self.public_description_html: str = d['public_description_html']
        self.sidebar_description: str = d['description']
        self.sidebar_description_html: str = d['description_html']
        self.submitting_form_note: str = d['submit_text']
        self.submitting_form_note_html: str = d['submit_text_html']
        self.submit_text_label: str = d['submit_text_label']
        self.submit_link_label: str = d['submit_link_label']

        submission_type: str = d['submission_type']
        self.allows_text_submissions: bool = submission_type in ('any', 'self')
        self.allows_link_submissions: bool = submission_type in ('any', 'link')
        self.allows_gallery_submissions: bool = d['allow_galleries']
        self.allows_poll_submissions: bool = d['allow_polls']

        self.suggested_comment_sort: Optional[str] = d['suggested_comment_sort']

        self.nsfw: bool = d['over18']
        self.public_traffic: bool = d['public_traffic']
        self.quarantined: bool = d['quarantine']

        self.me: Optional[BaseSubreddit.Me] = None
        if d['user_is_moderator'] is not None:
            self.me = self.Me(d)

        self.flair: BaseSubreddit.SubredditFlair = self.SubredditFlair(d)

        # Value is always `False` if object from search item.
        self.has_menu_widget: bool = d['has_menu_widget']


class BaseInaccessibleSubreddit(Artifact):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)

        self.name: str = d['display_name']
        self.openness: str = d['subreddit_type']

        self.title: str = d['title']




from __future__ import annotations
from typing import Sequence

from ..listing.listing_async_paginator import ListingAsyncPaginator
from ....model_loaders.subreddit_ASYNC import load_subreddit
from ....models.subreddit_ASYNC import Subreddit

class SubredditListingAsyncPaginator(ListingAsyncPaginator[Subreddit]):
    async def fetch(self) -> Sequence[Subreddit]:
        data = await self._fetch_data()
        return [load_subreddit(d['data'], self.client) for d in data['children']]




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Optional, Mapping, Any, Union
if TYPE_CHECKING:
    from ..http.http_client_SYNC import HTTPClient
    from ..http.payload import RequestFiles

from .core.http_client_SYNC import build_http_client
from ..http.util.json_load import json_loads_response

class Client:
    _TSelf = TypeVar('_TSelf', bound='Client')

    @classmethod
    def from_http(cls: type[_TSelf], http: HTTPClient) -> _TSelf:
        self = cls.__new__(cls)
        self._init(http)
        return self

    def __init__(self) -> None:
        http = build_http_client()
        self._init(http)

    def _init(self, http: HTTPClient) -> None:
        self.http: HTTPClient = http

        from .siteprocs.SYNC import Procedures
        self.p: Procedures = Procedures(self)

    def request(self,
        verb: str,
        uri: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        data: Optional[Union[Mapping[str, str], bytes]] = None,
        json: Any = None,
        files: Optional[RequestFiles] = None,
        timeout: float = -2,
    ) -> Any:
        json_data = None
        resp = self.http.request(verb, uri, params=params, headers=headers,
                data=data, json=json, files=files, timeout=timeout)
        if resp.data:
            json_data = json_loads_response(resp)
        resp.raise_for_status()
        return json_data

PushshiftClient: type[Client] = Client
Pushshift: type[Client] = Client
PushshiftAPI: type[Client] = Client




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .....client_SYNC import Client
    from .....models.subreddit_SYNC import Subreddit
    from ..stream import IStandardStreamEventSubject

from ..stream import Stream


def make_user_subreddit_stream(client: Client) -> IStandardStreamEventSubject[Subreddit]:
    it = client.p.user.pull_user_subreddits.new()
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: x.id)




__all__ = '*'

from . import (SYNC, ASYNC)  # noqa: F401
from . import (client_SYNC, client_ASYNC)  # noqa: F401
from .__about__ import (  # noqa: F401
    __title__ as __title__,
    __summary__ as __summary__,
    __uri__ as __uri__,
    __version__ as __version__,
    __author__ as __author__,
    __license__ as __license__,
    __copyright__ as __copyright__,
)
from . import auth  # noqa: F401
from . import core  # noqa: F401
from . import http  # noqa: F401
from . import websocket  # noqa: F401
from . import iterators  # noqa: F401
from . import pagination  # noqa: F401
from . import models  # noqa: F401
from . import dto_models  # noqa: F401
from . import model_loaders  # noqa: F401
from . import util  # noqa: F401
from . import exceptions  # noqa: F401



import sys

globals()[__name__] = __import__(__name__)

from . import _cli
sys.modules[__name__ + '.cli'] = _cli




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Sequence
if TYPE_CHECKING:
    from .menu_bar import Tab

from dataclasses import dataclass

@dataclass(repr=False, eq=False)
class MenuBar:
    d: Mapping[str, Any]
    idt: str
    kind: str
    show_wiki: bool
    tabs: Sequence[Tab]




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .user_base import (
    BaseUser,
    BaseSuspendedUser,
)

class User(BaseUser):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

class SuspendedUser(BaseSuspendedUser):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Sequence
if TYPE_CHECKING:
    from ..client_SYNC import Client

from .submission_collection_base import (
    BaseSubmissionCollectionDetails,
    GBaseSubmissionCollection,
)
from .submission_SYNC import Submission
from ..model_loaders.submission_SYNC import load_submission

class SubmissionCollectionDetails(BaseSubmissionCollectionDetails):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

    def add_post(self, submission_id: int) -> None:
        self.client.p.collection.add_post(self.uuid, submission_id)

    def remove_post(self, submission_id: int) -> None:
        self.client.p.collection.remove_post(self.uuid, submission_id)

    def reorder(self, submission_ids: Sequence[int]) -> None:
        self.client.p.collection.reorder(self.uuid, submission_ids)

    def delete(self) -> None:
        self.client.p.collection.delete(self.uuid)

    def set_title(self, title: str) -> None:
        self.client.p.collection.set_title(self.uuid, title)

    def set_description(self, desc: str) -> None:
        self.client.p.collection.set_description(self.uuid, desc)


class SubmissionCollection(
    SubmissionCollectionDetails,
    GBaseSubmissionCollection[Submission],
):
    def _load_submission(self, m: Mapping[str, Any]) -> Submission:
        return load_submission(m, self.client)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_user_item import (
        ModeratorUserItem,
        ContributorUserItem,
        BannedUserItem,
        MutedUserItem,
    )

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.async1 import (
    ModeratorsAsyncPaginator,
    ContributorsAsyncPaginator,
    BannedAsyncPaginator,
    MutedAsyncPaginator,
)

class PullUsers:
    def __init__(self, client: Client) -> None:
        self._client = client

    def moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ModeratorsAsyncPaginator, ModeratorUserItem]:
        p = ModeratorsAsyncPaginator(self._client, f'/api/v1/{sr}/moderators')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def moderator_invitations(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ModeratorsAsyncPaginator, ModeratorUserItem]:
        p = ModeratorsAsyncPaginator(self._client, f'/api/v1/{sr}/moderators_invited')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def editable_moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ModeratorsAsyncPaginator, ModeratorUserItem]:
        p = ModeratorsAsyncPaginator(self._client, f'/api/v1/{sr}/moderators_editable')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def contributors(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[ContributorsAsyncPaginator, ContributorUserItem]:
        p = ContributorsAsyncPaginator(self._client, f'/api/v1/{sr}/contributors')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def banned(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[BannedAsyncPaginator, BannedUserItem]:
        p = BannedAsyncPaginator(self._client, f'/api/v1/{sr}/banned')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def muted(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingAsyncIterator[MutedAsyncPaginator, MutedUserItem]:
        p = MutedAsyncPaginator(self._client, f'/api/v1/{sr}/muted')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable
if TYPE_CHECKING:
    from ...client_ASYNC import Client

from .listing.mixins.sort_ASYNC import Sort
from .listing.subreddit_listing_async_paginator import SubredditListingAsyncPaginator
from ...models.subreddit_ASYNC import Subreddit

class SearchSubredditsAsyncPaginator(
    Sort[Subreddit],
    SubredditListingAsyncPaginator,
):
    def __init__(self,
        client: Client,
        uri: str,
        query: str,
        show_users: bool = False,
    ):
        super().__init__(client, uri)
        self.query: str = query
        self.show_users: bool = show_users

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        yield ('q', self.query)
        if self.show_users:
            yield ('show_users', '1')




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, MutableSequence, Callable, Generic, TypeVar, Protocol, Awaitable, Generator, Optional, Iterator, AsyncIterator
if TYPE_CHECKING:
    from asyncio import Future
    from ....pagination.async_paginator import CursorAsyncPaginator

import asyncio
import random
import time

from ....pagination.async_paginator import Resettable
from ....util.ordered_set import BoundedSet
from ...util.event_dispatcher import EventDispatcher


TEvent_contra = TypeVar('TEvent_contra', contravariant=True)
TEvent = TypeVar('TEvent')

class StreamEventHandlerProtocol(Protocol[TEvent_contra]):
    async def __call__(self, event: TEvent_contra, /) -> None:
        ...

class StreamEventDispatcher(EventDispatcher[StreamEventHandlerProtocol[TEvent]], Generic[TEvent]):
    async def __call__(self, event: TEvent) -> None:
        for handler in self:
            await handler(event)


TOutput = TypeVar('TOutput')

class IStandardStreamEventSubject(Awaitable[None], AsyncIterator[float], Generic[TOutput]):
    output: StreamEventDispatcher[TOutput]
    error: StreamEventDispatcher[Exception]

    def __aiter__(self) -> AsyncIterator[float]: ...
    async def __anext__(self) -> float: ...
    def __await__(self) -> Generator[Optional[Future[None]], None, None]: ...


class Stream(IStandardStreamEventSubject[TOutput]):
    _BASE_POLL_INTERVAL: float = 5.0
    _MAX_POLL_INTERVAL: float = 50.0
    _BACKOFF_FACTOR: float =  2.0
    _JITTER_FACTOR: float = 0.4

    _MEMORY: int = 2000
    _TARGET_LIMIT_MULTIPLIER: float = 1.2

    _BACKTRACK_DEPTH: int = 300

    def __init__(self,
        paginator: CursorAsyncPaginator[TOutput],
        extractor: Callable[[TOutput], object],
        *,
        max_limit: int = 100,
    ) -> None:
        if not isinstance(paginator, Resettable):
            raise ValueError('paginator must be Resettable')
        self._paginator: CursorAsyncPaginator[TOutput] = paginator
        self._paginator__resettable: Resettable = paginator
        self._extractor: Callable[[TOutput], object] = extractor
        self._max_limit: int = max_limit

        self._agen: AsyncIterator[float] = self.__routine()

        self.output: StreamEventDispatcher[TOutput] = StreamEventDispatcher()
        self.error: StreamEventDispatcher[Exception] = StreamEventDispatcher()

    def __aiter__(self) -> AsyncIterator[float]:
        return self

    async def __anext__(self) -> float:
        return await self._agen.__anext__()

    def __await__(self) -> Generator[Optional[Future[None]], None, None]:
        async def _coro_func() -> None:
            async for t in self:
                await asyncio.sleep(t)
        return _coro_func().__await__()

    @staticmethod
    def _intermit(duration: float) -> Iterator[float]:
        point = time.monotonic() + duration
        yield duration
        while (n := point - time.monotonic()) > 0:
            yield n

    async def __routine(self) -> AsyncIterator[float]:
        paginator = self._paginator
        paginator__resettable = self._paginator__resettable
        extractor = self._extractor
        max_limit = self._max_limit

        seen: BoundedSet[object] = BoundedSet((), self._MEMORY)
        delay: float = self._BASE_POLL_INTERVAL

        paginator.limit = max_limit

        retrieved: Sequence[TOutput] = ()
        while not retrieved:
            try:
                retrieved = await paginator.fetch()
            except Exception as error:
                await self.emit_error(error)
                if delay < self._MAX_POLL_INTERVAL:
                    delay = min(self._BACKOFF_FACTOR * delay, self._MAX_POLL_INTERVAL)
                for v in self._intermit(random.uniform(
                        delay * (1 - self._JITTER_FACTOR),
                        delay * (1 + self._JITTER_FACTOR))):
                    yield v

        delay = self._BASE_POLL_INTERVAL

        paginator__resettable.reset()

        for obj in retrieved:
            seen.add(extractor(obj))

        yield 0

        tuna_limit: float = max_limit
        backtrack_count: int = 0

        while True:
            effective_limit = round(tuna_limit)
            paginator.limit = effective_limit

            retrieved = ()
            fetch_successful: bool = True
            try:
                retrieved = await paginator.fetch()
            except Exception as error:
                fetch_successful = False
                await self.emit_error(error)

            selection: MutableSequence[TOutput] = []
            for obj in retrieved:
                identity = self._extractor(obj)
                if identity not in seen:
                    seen.add(identity)
                    selection.append(obj)

            for obj in selection:
                await self.emit_output(obj)

            backtrack_count += len(selection)

            if (len(selection) < effective_limit) or (backtrack_count > self._BACKTRACK_DEPTH):
                paginator__resettable.reset()
                backtrack_count = 0

            if fetch_successful:
                tuna = tuna_limit
                if not selection:
                    tuna = max(tuna/2, 1)
                elif len(selection) >= effective_limit:
                    tuna = min(2*tuna, max_limit)
                else:
                    tuna = min((self._TARGET_LIMIT_MULTIPLIER * len(selection)), max_limit)
                tuna_limit = tuna

            if fetch_successful:
                delay = self._BASE_POLL_INTERVAL
            else:
                if delay < self._MAX_POLL_INTERVAL:
                    delay = min(self._BACKOFF_FACTOR * delay, self._MAX_POLL_INTERVAL)

            t = 1.
            if not paginator.get_cursor():
                t = random.uniform(
                        delay * (1 - self._JITTER_FACTOR),
                        delay * (1 + self._JITTER_FACTOR))
            for v in self._intermit(t):
                yield v

    async def emit_output(self, output: TOutput) -> None:
        await self.output(output)

    async def emit_error(self, error: Exception) -> None:
        await self.error(error)




from __future__ import annotations
from typing import Callable

from collections.abc import Sized
import time
from heapq import heappush, heappop

class Entry:
    def __init__(self, when: float, callback: Callable[[], None]) -> None:
        self.when: float = when
        self.callback: Callable[[], None] = callback
        self.scheduled: bool = True
        self.cancelled: bool = False

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when == other.when
        return NotImplemented
    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when < other.when
        return NotImplemented
    def __le__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when <= other.when
        return NotImplemented
    def __gt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when > other.when
        return NotImplemented
    def __ge__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when >= other.when
        return NotImplemented

class Handle:
    def __init__(self, entry: Entry) -> None:
        self._entry = entry

    def get_when(self) -> float:
        return self._entry.when

    def get_cancelled(self) -> bool:
        return self._entry.cancelled

    def get_scheduled(self) -> bool:
        return self._entry.scheduled

    def cancel(self) -> None:
        self._entry.cancelled = True


class Scheduler(Sized):
    def __init__(self,
        *,
        time_func: Callable[[], float] = time.monotonic,
        delay_func: Callable[[float], None] = time.sleep,
    ) -> None:
        self.time_func: Callable[[], float] = time_func
        self.delay_func: Callable[[float], None] = delay_func
        self._pq: list[Entry] = []

    def __len__(self) -> int:
        return len(self._pq)

    def empty(self) -> bool:
        return not self._pq

    def time(self) -> float:
        return self.time_func()

    def call_at(self, when: float, callback: Callable[[], None]) -> Handle:
        entry = Entry(when, callback)
        heappush(self._pq, entry)
        return Handle(entry)

    def call_later(self, delay: float, callback: Callable[[], None]) -> Handle:
        return self.call_at(self.time() + delay, callback)

    def call_soon(self, callback: Callable[[], None]) -> Handle:
        return self.call_later(0, callback)

    def jog(self) -> float:
        pq = self._pq
        while pq:
            entry = pq[0]
            if entry.cancelled:
                heappop(pq)
                entry.scheduled = False
                continue

            curr_time = self.time_func()
            if entry.when > curr_time:
                return entry.when - curr_time

            heappop(pq)
            entry.scheduled = False
            entry.callback()
        return -1

    def run(self) -> None:
        while True:
            t = self.jog()
            if t < 0:
                break
            self.delay_func(t)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...model_loaders.subreddit_SYNC import load_potentially_inaccessible_subreddit
from ...util.base_conversion import to_base36

class Get:
    def __init__(self, client: Client):
        self._client = client

    def __call__(self, id: int) -> Optional[object]:
        id36 = to_base36(id)
        return self.by_id36(id36)

    def by_id36(self, id36: str) -> Optional[object]:
        full_id36 = 't5_' + id36
        root = self._client.request('GET', '/api/info', params={'id': full_id36})
        if children := root['data']['children']:
            return load_potentially_inaccessible_subreddit(children[0]['data'], self._client)
        return None




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .request import Request
    from .response import Response

from .requestor_ASYNC import Requestor

class RequestorAugmenter(Requestor):
    def __init__(self, requestor: Requestor) -> None:
        self.requestor: Requestor = requestor

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        return await self.requestor.send(request, timeout=timeout)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterator, Optional
if TYPE_CHECKING:
    from ....client_SYNC import Client

from ...util.sched import Scheduler


def run(*streams: Iterator[float], client: Optional[Client] = None) -> None:
    run_series(*streams, client=client)


def run_series(*streams: Iterator[float], client: Optional[Client] = None) -> None:
    def _invoke_and_reschedule(itr: Iterator[float], shdlr: Scheduler) -> None:
        try:
            t = next(itr)
        except StopIteration:
            pass
        else:
            shdlr.call_later(t, (lambda: _invoke_and_reschedule(itr, shdlr)))

    shdlr = Scheduler()
    for itr in streams:
        _invoke_and_reschedule(itr, shdlr)

    try:
        shdlr.run()

    finally:
        if client is not None:
            client.close()




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...util.base_conversion import to_base36

class AddPost:
    def __init__(self, client: Client):
        self._client = client

    def __call__(self, uuid: str, submission_id: int) -> None:
        id36 = to_base36(submission_id)
        return self.by_id36(uuid, id36)

    def by_id36(self, uuid: str, submission_id36: str) -> None:
        params = {'collection_id': uuid, 'link_fullname': 't3_' + submission_id36}
        self._client.request('POST', '/api/v1/collections/add_post_to_collection', params=params)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ...http.requestor_SYNC import Requestor
    from ...http.request import Request
    from ...http.response import Response

import time

from ...http.requestor_augmenter_SYNC import RequestorAugmenter
from ...util.token_bucket import TokenBucket

class RateLimited(RequestorAugmenter):
    def __init__(self, requestor: Requestor) -> None:
        super().__init__(requestor)
        self._tb = TokenBucket(3, 1/2)

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        time.sleep(self._tb.get_cooldown(1))
        self._tb.consume(1)
        return self.requestor.send(request, timeout=timeout)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_SYNC import ExtraSubmissionFieldsComment

from .pull_SYNC import Pull

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.listing.comment_listing_paginator import ExtraSubmissionFieldsCommentListingPaginator

class FrontProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.pull: Pull = Pull(client)

    def pull_new_comments(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ExtraSubmissionFieldsCommentListingPaginator, ExtraSubmissionFieldsComment]:
        p = ExtraSubmissionFieldsCommentListingPaginator(self._client, '/comments')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations

import sys
from os import path as op
from os import getenv
from configparser import ConfigParser

def safe_getenv(key: str) -> str:
    return getenv(key, '')

def get_praw_ini_locations() -> list[str]:
    pkg_dir = ''
    if __name__ != '__main__':
        root_pkg_name, _, _ = __name__.partition('.')
        root_pkg_module = sys.modules[root_pkg_name]
        if root_pkg_module.__file__:
            pkg_dir = op.dirname(root_pkg_module.__file__)

    return [
        op.join(*path_components, 'praw.ini')
        for path_components in [
            (pkg_dir,),  # Package defaults
            (safe_getenv('APPDATA'),),  # Windows
            (safe_getenv('HOME'), '.config',),  # Linux and macOS
            (safe_getenv('XDG_CONFIG_HOME'),),  # Linux
            ('.',),  # Current directory
        ]
        if path_components[0]
    ]

def get_praw_config() -> ConfigParser:
    config = ConfigParser()
    config.read(get_praw_ini_locations())
    return config




from __future__ import annotations
from typing import Any, Mapping, Optional

from functools import cached_property
from datetime import datetime, timezone

class UserRelationshipItem:
    @cached_property
    def added_at(self) -> datetime:
        return datetime.fromtimestamp(self.added_ut, timezone.utc)

    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        self.id: int = int(id36, 36)
        self.name: str = d['name']
        #self.rel_id: str = d['rel_id']
        self.added_ut: int = int(d['date'])

class FriendRelationshipItem(UserRelationshipItem):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        # Need premium to test this
        #self.note: Optional[str] = d['note']

class BannedUserRelationshipItem(UserRelationshipItem):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.days_remaining: Optional[int] = d['days_left']
        self.detail: str = d['note']




from __future__ import annotations
from typing import Iterable, Iterator, Callable, TypeVar, Awaitable

from .stubborn_caller_async_iterator import StubbornCallerAsyncIterator

T = TypeVar('T')

class CallChunkCallingAsyncIterator(StubbornCallerAsyncIterator[T]):
    """Evaluate call chunks and return their results."""

    def __init__(self, chunks: Iterable[Callable[[], Awaitable[T]]]) -> None:
        self.__chunk_iter: Iterator[Callable[[], Awaitable[T]]] = iter(chunks)
        super().__init__(self.__chunk_iter)

    def get_chunk_iter(self) -> Iterator[Callable[[], Awaitable[T]]]:
        return self.__chunk_iter




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Mapping, MutableMapping, Union, Any
if TYPE_CHECKING:
    from .payload import Payload, RequestFiles

from dataclasses import dataclass

from .payload import make_payload
from .util.case_insensitive_dict import CaseInsensitiveDict

@dataclass(eq=False, repr=False)
class Request:
    verb: str
    uri: str
    params: MutableMapping[str, str]
    headers: MutableMapping[str, str]
    payload: Optional[Payload]

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} [{self.verb}]>'

def make_request(
    verb: str,
    uri: str,
    *,
    params: Optional[Mapping[str, str]] = None,
    headers: Optional[Mapping[str, str]] = None,
    data: Optional[Union[Mapping[str, str], bytes]] = None,
    json: Any = None,
    files: Optional[RequestFiles] = None,
) -> Request:
    params = dict(params or {})
    headers = CaseInsensitiveDict(headers or {})
    payload = make_payload(data, json, files)
    return Request(verb, uri, params=params, headers=headers, payload=payload)




from .reddit_http_client_ASYNC import RedditHTTPClient as RedditHTTPClient  # noqa: F401
from .reddit_token_obtainment_client_ASYNC import RedditTokenObtainmentClient as RedditTokenObtainmentClient  # noqa: F401
from .authorizer_ASYNC import Authorizer as Authorizer, Authorized as Authorized  # noqa: F401
from .rate_limited_ASYNC import RateLimited as RateLimited  # noqa: F401




from __future__ import annotations
from typing import Any, Dict, Collection, TypeVar, Iterator, Iterable, Tuple, Mapping, IO, cast, Generic, Callable

import inspect
from ast import literal_eval
import reprlib
from pprint import PrettyPrinter
from io import StringIO


rep: reprlib.Repr = reprlib.Repr()
rep.maxlevel = 1
rep.maxstring = 250
rep.maxother = 250
reprepr: Callable[[Any], str] = rep.repr

class StrReprStr(str):
    def __repr__(self) -> str:
        return str(self)

def neat_repr_dict(d: Dict[Any, Any]) -> Dict[Any, Any]:
    return {
        k: (literal_eval(reprepr(v)) if isinstance(v, str) else StrReprStr(reprepr(v)))
        for k, v in d.items()
    }


ppr: PrettyPrinter = PrettyPrinter()
ppr_format: Any = ppr._format  # type: ignore[attr-defined]

def pretty_format(obj: object) -> str:
    # This function ensures `obj`'s pretty-print-formatted string is
    # returned, since `pprint.pformat()` only invokes an object's
    # pretty printer if its __repr__ is long enough.
    sio = StringIO()
    ppr_format(obj, sio, 0, 999999, {}, 0)
    return sio.getvalue()


T = TypeVar('T')

class DataMembersNamespace(Collection[str], Generic[T]):
    def __init__(self, instance: T):
        self._instance = instance

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}({self._instance})>'

    def __str__(self) -> str:
        return pretty_format(self)

    def __iter__(self) -> Iterator[str]:
        return (k for k, _v in self._data_members())

    def __len__(self) -> int:
        return sum(1 for _ in self._data_members())

    def __contains__(self, item: object) -> bool:
        try:
            value = getattr(self._instance, cast(str, item))
        except AttributeError:
            return False
        return self._data_member(value)

    def __getattr__(self, name: str) -> Any:
        value = getattr(self._instance, name)
        if not self._data_member(value):
            raise AttributeError(f'{self._instance.__class__.__name__!r} has no data attribute {name!r}')
        return value

    def __dir__(self) -> Iterable[str]:
        return list(self)

    def __abs__(self) -> T:
        return self._instance

    def _data_member(self, value: object) -> bool:
        return not callable(value)

    def _data_members(self) -> Iterator[Tuple[str, Any]]:
        for name, value in inspect.getmembers(self._instance, self._data_member):
            if value is self:
                continue
            if name.startswith('_'):
                continue
            yield name, value

    @staticmethod
    def _pprint(
        printer: PrettyPrinter,
        obj: DataMembersNamespace[T],
        stream: IO[str],
        indent: int,
        allowance: int,
        context: Mapping[int, Any],
        level: int,
    ) -> None:
        leader = 'data members: '
        stream.write(leader)
        printer._format(  # type: ignore[attr-defined]
            neat_repr_dict(dict(obj._data_members())),
            stream,
            indent + len(leader),
            allowance,
            context,
            level,
        )

    if isinstance(getattr(PrettyPrinter, '_dispatch', None), dict):
        PrettyPrinter._dispatch[__repr__] = _pprint.__func__  # type: ignore[attr-defined]


class DataMembersNamespaceMapping(DataMembersNamespace[T]):
    def __getitem__(self, key: str) -> Any:
        try:
            value = getattr(self._instance, key)
        except AttributeError:
            raise KeyError(key)
        if not self._data_member(value):
            raise KeyError(key)
        return value




from typing import Any, Mapping, Optional, NamedTuple, Sequence

from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class FlairTemplate(IArtifact):
    d: Mapping[str, Any]
    uuid: str
    type: str
    mod_only: bool
    text: str
    bg_color: str
    fg_light_or_dark: str
    allowable_content: str
    css_class: str
    max_emojis: int
    text_editable: bool

@dataclass(repr=False, eq=False)
class CurrentFlairChoice:
    uuid: Optional[str]
    text: str
    css_class: str
    position: str

@dataclass(repr=False, eq=False)
class FlairTemplateChoice:
    uuid: str
    text: str
    css_class: str
    position: str
    text_editable: bool

class FlairChoices(NamedTuple):
    current: Optional[CurrentFlairChoice]
    choices: Sequence[FlairTemplateChoice]

@dataclass(repr=False, eq=False)
class UserFlairAssociation:
    username: str
    text: str
    has_had_css_class_when_no_flair_template: bool
    css_class: str




from . import (SYNC, ASYNC)  # noqa: F401
from . import (reddit_http_client_SYNC, reddit_http_client_ASYNC)  # noqa: F401
from . import (reddit_token_obtainment_client_SYNC, reddit_token_obtainment_client_ASYNC)  # noqa: F401
from . import (authorizer_SYNC, authorizer_ASYNC)  # noqa: F401
from . import (rate_limited_SYNC, rate_limited_ASYNC)  # noqa: F401
from . import exceptions  # noqa: F401




from typing import Mapping, Any

from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class MediaUploadLease(IArtifact):
    d: Mapping[str, Any]
    endpoint: str
    fields: Mapping[str, str]
    s3_object_key: str
    location: str
    media_id: str
    declared_filename: str
    websocket_url: str




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.subreddit_user_item import (
        ModeratorUserItem,
        ContributorUserItem,
        BannedUserItem,
        MutedUserItem,
    )

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.sync1 import (
    ModeratorsPaginator,
    ContributorsPaginator,
    BannedPaginator,
    MutedPaginator,
)

class PullUsers:
    def __init__(self, client: Client) -> None:
        self._client = client

    def moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, ModeratorUserItem]:
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators')
        return ImpartedPaginatorChainingIterator(p, amount)

    def moderator_invitations(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, ModeratorUserItem]:
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators_invited')
        return ImpartedPaginatorChainingIterator(p, amount)

    def editable_moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, ModeratorUserItem]:
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators_editable')
        return ImpartedPaginatorChainingIterator(p, amount)

    def contributors(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ContributorsPaginator, ContributorUserItem]:
        p = ContributorsPaginator(self._client, f'/api/v1/{sr}/contributors')
        return ImpartedPaginatorChainingIterator(p, amount)

    def banned(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[BannedPaginator, BannedUserItem]:
        p = BannedPaginator(self._client, f'/api/v1/{sr}/banned')
        return ImpartedPaginatorChainingIterator(p, amount)

    def muted(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[MutedPaginator, MutedUserItem]:
        p = MutedPaginator(self._client, f'/api/v1/{sr}/muted')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import Mapping, Any, Optional

from datetime import datetime, timezone

from ..auth.const import AUTHORIZATION_BASE_URL
from .artifact import Artifact

class BaseMailboxMessage(Artifact):
    pass

class BaseComposedMessage(BaseMailboxMessage):
    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id: int = int(d['id'], 36)
        self.timestamp: int = int(d['created_utc'])
        self.datetime: datetime = datetime.fromtimestamp(self.timestamp, timezone.utc)
        dest: str = d['dest']
        self.dest: str = dest
        self.prefixed_recipient: str = '%s/%s' % ('ur'[dest.startswith('#')], dest)
        self.subject: str = d['subject']
        self.body: str = d['body']
        self.body_html: str = d['body_html']
        self.unread: bool = d['new']
        self.distinguished: str = d['distinguished'] or ''
        self.author_name: Optional[str] = d['author']
        self.author_id: Optional[int] = (
            int(v.partition('_')[2], 36)
            if (v := d['author_fullname']) else
            None
        )
        self.via: Optional[str] = None if dest.startswith('#') else d['subreddit']

class BaseCommentMessage(BaseMailboxMessage):
    class SubmissionInfo:
        def __init__(self, d: Mapping[str, Any]):
            self.title: str = d['link_title']
            context: str = d['context']
            self.id: int = int(context.split('/', 5)[4], 36)
            self.comment_count: int = d['num_comments']

    class CommentInfo:
        def __init__(self, d: Mapping[str, Any]):
            self.id: int = int(d['id'], 36)
            self.timestamp: int = int(d['created_utc'])
            self.datetime: datetime = datetime.fromtimestamp(self.timestamp, timezone.utc)
            self.context: str = d['context']
            self.rel_permalink: str = self.context.partition('?')[0]
            self.permalink: str = AUTHORIZATION_BASE_URL + self.rel_permalink

            parent_id: str = d['parent_id']
            self.is_top_level: bool = parent_id.startswith('t3_')
            self.parent_comment_id36: Optional[str] = None
            self.parent_comment_id: Optional[int] = None
            if parent_id.startswith('t1_'):
                self.parent_comment_id36 = parent_id.partition('_')[2]
                self.parent_comment_id = int(self.parent_comment_id36, 36)

            self.score: int = d['score']
            self.subreddit_name: str = d['subreddit']
            self.body: str = d['body']
            self.body_html: str = d['body_html']
            self.author_name: str = d['author']
            self.author_id: int = int(d['author_fullname'].partition('_')[2], 36)

            self.voted: int = {False: -1, None: 0, True: 1}[d['likes']]

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.type: str = d['type']
        self.submission: BaseCommentMessage.SubmissionInfo = self.SubmissionInfo(d)
        self.comment: BaseCommentMessage.CommentInfo = self.CommentInfo(d)




from __future__ import annotations
from typing import TYPE_CHECKING, Any
if TYPE_CHECKING:
    from ..http.request import Request
    from ..http.response import Response

from .. import auth
from ..auth.const import TOKEN_OBTAINMENT_URL
from .. import http
from ..http.util.case_insensitive_dict import CaseInsensitiveDict
from ..exceptions import ArgExcMixin
from ..auth.exceptions import raise_for_token_server_response_error

class ArgExc(ArgExcMixin):
    pass


def raise_for_reddit_token_server_response_error(json_dict: Any) -> None:
    error = json_dict.get('error')
    if not isinstance(error, str):
        return
    raise_for_token_server_response_error(json_dict)


class _Throwaway(ArgExc):
    pass

class AuthError(_Throwaway):
    pass

class CredentialsError(_Throwaway):
    pass

class ClientCredentialsError(CredentialsError):
    pass

class GrantCredentialsError(CredentialsError):
    pass

class BlacklistedUserAgent(_Throwaway):
    pass

class FaultyUserAgent(_Throwaway):
    pass


def raise_for_reddit_auth_response_exception(e: Exception, req: Request, resp: Response) -> None:
    req_headers = CaseInsensitiveDict(req.headers)

    if isinstance(e, auth.exceptions.TokenServerResponseError):
        if e.error_name == 'invalid_grant':
            msg = 'Check your grant credentials'
            if isinstance(pld := req.payload, http.payload.URLEncodedFormData):
                grant = pld.data
                if grant.get('grant_type') == 'password':
                    msg = "Check your grant credentials.\n\nTip: If you have 2FA enabled the password grant won't work. You must use a refresh token instead."
            raise GrantCredentialsError(msg)

        elif e.error_name == 'unsupported_grant_type':
            content_type = req_headers.get('Content-Type', '')
            expected_content_type = 'application/x-www-form-urlencoded'
            if content_type != expected_content_type:
                raise AuthError(f'bad Content-Type header: got {content_type!r}, need {expected_content_type!r}')

    elif isinstance(e, http.exceptions.StatusCodeException):
        status = e.status_code
        if status == 400:
            if isinstance(pld := req.payload, http.payload.URLEncodedFormData):
                grant = pld.data
                if grant.get('grant_type') == 'refresh_token':
                    e.arg = "Your refresh token might be invalid."
                    raise e

        elif status == 401:
            if not (url := req.uri).startswith(("https://www.reddit.com", "https://ssl.reddit.com")):
                raise AuthError(f'bad access token URL: got {url!r}, need {TOKEN_OBTAINMENT_URL!r}')
            if 'Authorization' not in req_headers:
                raise AuthError('Authorization header missing from request')
            if not req_headers['Authorization'][:5].lower().startswith('basic'):
                raise AuthError('Authorization header value must start with "Basic"')
            raise ClientCredentialsError('Check your client credentials')

        elif status == 403:
            ua = req_headers['User-Agent']
            if ua.startswith('Bot'):
                raise BlacklistedUserAgent("User-Agent strings must not start with 'Bot'.")
            for s in ['scraping', 'searchme']:
                if s in ua:
                    raise BlacklistedUserAgent(
                        f"{s!r} is a known blacklisted User-Agent pattern."
                        " Remove it from your User-Agent string."
                    )

        elif status == 429:
            ua = req_headers['User-Agent']
            if 'curl' in ua:
                msg = "The pattern 'curl', in your User-Agent string, is known to interfere with rate limits. Remove it from your User-Agent string."
                raise FaultyUserAgent(msg)

        elif status == 500:
            if isinstance(pld := req.payload, http.payload.URLEncodedFormData):
                grant = pld.data
                if grant.get('grant_type') == 'authorization_code':
                    e.arg = "The authorization code might be expired."
                    raise e




from __future__ import annotations
from typing import Any, Mapping

from dataclasses import dataclass

from .artifact import IArtifact

@dataclass(repr=False, eq=False)
class Trophy(IArtifact):
    d: Mapping[str, Any]
    name: str
    description: str
    icon_40: str
    icon_70: str




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .....client_ASYNC import Client
    from .....models.modmail_ASYNC import Conversation, Message
    from ..stream import IStandardStreamEventSubject

from ..stream import Stream


def make_conversation_and_message_stream(client: Client) -> IStandardStreamEventSubject[tuple[Conversation, Message]]:
    it = client.p.modmail.conversation.pulls.new()
    paginator = it.get_paginator()
    return Stream(paginator, lambda x: (x[0].id, x[1].id))




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Iterable
if TYPE_CHECKING:
    from ..client_ASYNC import Client
    from ..iterators.call_chunk_calling_async_iterator import CallChunkCallingAsyncIterator

from .custom_feed_base import BaseCustomFeed

class CustomFeed(BaseCustomFeed):
    def __init__(self, d: Mapping[str, Any], client: Client):
        super().__init__(d)
        self.client: Client = client

    async def delete(self) -> None:
        await self.client.p.custom_feed.delete(self.owner, self.name)

    async def contains(self, sr_name: str) -> bool:
        return await self.client.p.custom_feed.contains(self.owner, self.name, sr_name)

    async def add_to(self, sr_name: str) -> None:
        await self.client.p.custom_feed.add_to(self.owner, self.name, sr_name)

    def bulk_add_to(self, sr_names: Iterable[str]) -> CallChunkCallingAsyncIterator[None]:
        return self.client.p.custom_feed.bulk_add_to(self.owner, self.name, sr_names)

    async def remove_from(self, sr_name: str) -> None:
        await self.client.p.custom_feed.remove_from(self.owner, self.name, sr_name)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Mapping
if TYPE_CHECKING:
    from ..requestor_SYNC import Requestor
    from ..request import Request
    from ..response import Response

from ..requestor_augmenter_SYNC import RequestorAugmenter

class ApplyParamsAndHeaders(RequestorAugmenter):
    def __init__(self, requestor: Requestor, *,
            params: Optional[Mapping[str, str]] = None,
            headers: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.params: Mapping[str, str] = {} if params is None else params
        self.headers: Mapping[str, str] = {} if headers is None else headers

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        request.params.update(self.params)
        request.headers.update(self.headers)
        return self.requestor.send(request, timeout=timeout)

class ApplyDefaultParamsAndHeaders(RequestorAugmenter):
    def __init__(self, requestor: Requestor, *,
            params: Optional[Mapping[str, str]] = None,
            headers: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.params: Mapping[str, str] = {} if params is None else params
        self.headers: Mapping[str, str] = {} if headers is None else headers

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        (pd := request.params).update({**self.params, **pd})
        (hd := request.headers).update({**self.headers, **hd})
        return self.requestor.send(request, timeout=timeout)



class ApplyParams(RequestorAugmenter):
    def __init__(self, requestor: Requestor, params: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.params: Mapping[str, str] = {} if params is None else params

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        request.params.update(self.params)
        return self.requestor.send(request, timeout=timeout)

class ApplyDefaultParams(RequestorAugmenter):
    def __init__(self, requestor: Requestor, params: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.params: Mapping[str, str] = {} if params is None else params

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        (pd := request.params).update({**self.params, **pd})
        return self.requestor.send(request, timeout=timeout)



class ApplyHeaders(RequestorAugmenter):
    def __init__(self, requestor: Requestor, headers: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.headers: Mapping[str, str] = {} if headers is None else headers

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        request.headers.update(self.headers)
        return self.requestor.send(request, timeout=timeout)

class ApplyDefaultHeaders(RequestorAugmenter):
    def __init__(self, requestor: Requestor, headers: Optional[Mapping[str, str]] = None):
        super().__init__(requestor)
        self.headers: Mapping[str, str] = {} if headers is None else headers

    def send(self, request: Request, *, timeout: float = -2) -> Response:
        (hd := request.headers).update({**self.headers, **hd})
        return self.requestor.send(request, timeout=timeout)




from __future__ import annotations
from typing import Mapping, Any, Sequence, Iterator, overload, Union

from datetime import datetime, timezone
from functools import cached_property

class Rule:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.kind: str = d['kind']
        self.description: str = d['description']
        self.description_html: str = d.get('description_html', '')
        self.short_name: str = d['short_name']
        self.violation_reason: str = d['violation_reason']
        self.created_ut: int = int(d['created_utc'])

    @cached_property
    def created_at(self) -> datetime:
        return datetime.fromtimestamp(self.created_ut, timezone.utc)


class SubredditRules(Sequence[Rule]):
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self._rules = [Rule(o) for o in d['rules']]

    def __len__(self) -> int:
        return len(self._rules)

    def __contains__(self, item: object) -> bool:
        return item in self._rules

    def __iter__(self) -> Iterator[Rule]:
        return iter(self._rules)

    @overload
    def __getitem__(self, index: int) -> Rule: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Rule]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Rule, Sequence[Rule]]:
        return self._rules[index]




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Sequence, TypeVar, Optional
if TYPE_CHECKING:
    from ..client_ASYNC import Client

import urllib.parse

from ..paginators.comment_search_async_paginator import CommentSearchDocumentAsyncPaginator, CommentSearchDocumentIDAsyncPaginator
from ..paginators.submission_search_async_paginator import SubmissionSearchDocumentAsyncPaginator, SubmissionSearchDocumentIDAsyncPaginator
from ..models.pushshift_document import PushshiftDocument
from ...util.base_conversion import to_base36
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...iterators.chunking import chunked

T = TypeVar('T')

class Procedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    def bulk_fetch_comments_by_id(self, ids: Iterable[int]) -> CallChunkChainingAsyncIterator[PushshiftDocument]:
        async def mass_fetch_comments_by_id(ids: Sequence[int]) -> Sequence[PushshiftDocument]:
            ids_str = ','.join(map(to_base36, ids))
            root = await self._client.request('GET', '/reddit/comment/search', params={'ids': ids_str})
            return [PushshiftDocument(d) for d in root['data']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch_comments_by_id, chunk) for chunk in chunked(ids, 100))

    def bulk_fetch_comments_by_id_selecting_fields(self, ids: Iterable[int], *, fields: Iterable[str]) -> CallChunkChainingAsyncIterator[PushshiftDocument]:
        async def mass_fetch_comments_by_id_selecting_fields(ids: Sequence[int]) -> Sequence[PushshiftDocument]:
            params = {
                'ids': ','.join(map(to_base36, ids)),
                'fields': ','.join(fields),
            }
            root = await self._client.request('GET', '/reddit/comment/search', params=params)
            return [PushshiftDocument(d) for d in root['data']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch_comments_by_id_selecting_fields, chunk) for chunk in chunked(ids, 100))



    def page_search_comments(self,
        query: str = '',
        *,
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
    ) -> CommentSearchDocumentAsyncPaginator:
        return CommentSearchDocumentAsyncPaginator(
            client=self._client,
            uri='/reddit/search/comment',
            query=query,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
        )

    def page_search_comments_select_id(self,
        query: str = '',
        *,
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
    ) -> CommentSearchDocumentIDAsyncPaginator:
        return CommentSearchDocumentIDAsyncPaginator(
            client=self._client,
            uri='/reddit/search/comment',
            query=query,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
        )

    def page_search_comments_selecting_fields(self,
        query: str = '',
        *,
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
        fields: Iterable[str],
    ) -> CommentSearchDocumentAsyncPaginator:
        return CommentSearchDocumentAsyncPaginator(
            client=self._client,
            uri='/reddit/search/comment',
            query=query,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
            fields=fields,
        )

    async def count_search_comments(self,
        query: str = '',
        *,
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
    ) -> int:
        p = self.page_search_comments(
            query=query,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
        )
        p.limit = 0
        query_pairs = list(p.generate_doseq_params()) + [('metadata', ('true',))]
        query_string = urllib.parse.urlencode(query_pairs, doseq=True)
        uri = '%s?%s' % (p.uri, query_string)
        root = await self._client.request('GET', uri)
        return root['metadata']['total_results']



    def page_search_submissions(self,
        query: str = '',
        *,
        query_exclude: str = '',
        query_title: str = '',
        query_title_exclude: str = '',
        query_body: str = '',
        query_body_exclude: str = '',
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
    ) -> SubmissionSearchDocumentAsyncPaginator:
        return SubmissionSearchDocumentAsyncPaginator(
            client=self._client,
            uri='/reddit/search/submission',
            query=query,
            query_exclude=query_exclude,
            query_title=query_title,
            query_title_exclude=query_title_exclude,
            query_body=query_body,
            query_body_exclude=query_body_exclude,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
        )

    def page_search_submissions_select_id(self,
        query: str = '',
        *,
        query_exclude: str = '',
        query_title: str = '',
        query_title_exclude: str = '',
        query_body: str = '',
        query_body_exclude: str = '',
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
    ) -> SubmissionSearchDocumentIDAsyncPaginator:
        return SubmissionSearchDocumentIDAsyncPaginator(
            client=self._client,
            uri='/reddit/search/submission',
            query=query,
            query_exclude=query_exclude,
            query_title=query_title,
            query_title_exclude=query_title_exclude,
            query_body=query_body,
            query_body_exclude=query_body_exclude,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
        )

    def page_search_submissions_selecting_fields(self,
        query: str = '',
        *,
        query_exclude: str = '',
        query_title: str = '',
        query_title_exclude: str = '',
        query_body: str = '',
        query_body_exclude: str = '',
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
        ascending: bool = False,
        fields: Iterable[str],
    ) -> SubmissionSearchDocumentAsyncPaginator:
        return SubmissionSearchDocumentAsyncPaginator(
            client=self._client,
            uri='/reddit/search/submission',
            query=query,
            query_exclude=query_exclude,
            query_title=query_title,
            query_title_exclude=query_title_exclude,
            query_body=query_body,
            query_body_exclude=query_body_exclude,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
            ascending=ascending,
            fields=fields,
        )

    async def count_search_submissions(self,
        query: str = '',
        *,
        query_exclude: str = '',
        query_title: str = '',
        query_title_exclude: str = '',
        query_body: str = '',
        query_body_exclude: str = '',
        author: str = '',
        subreddit: str = '',
        time_range: tuple[Optional[int], Optional[int]] = (None, None),
    ) -> int:
        p = self.page_search_submissions(
            query=query,
            query_exclude=query_exclude,
            query_title=query_title,
            query_title_exclude=query_title_exclude,
            query_body=query_body,
            query_body_exclude=query_body_exclude,
            author=author,
            subreddit=subreddit,
            time_range=time_range,
        )
        p.limit = 0
        query_pairs = list(p.generate_doseq_params()) + [('metadata', ('true',))]
        query_string = urllib.parse.urlencode(query_pairs, doseq=True)
        uri = '%s?%s' % (p.uri, query_string)
        root = await self._client.request('GET', uri)
        return root['metadata']['total_results']




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_user_item import (
        ModeratorUserItem,
        ContributorUserItem,
        BannedUserItem,
        MutedUserItem,
    )
    from ...models.moderation_action_log_entry import ModerationActionLogEntry

from functools import cached_property

from ...model_loaders.subreddit_user_item import (
    load_moderator_user_item,
    load_contributor_user_item,
    load_banned_user_item,
    load_muted_user_item,
)
from .pull_users_ASYNC import PullUsers
from .legacy_ASYNC import Legacy
from .pull_ASYNC import Pull
from .note_ASYNC import Note
from ...util.base_conversion import to_base36
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.moderation.async1 import ModerationActionLogAsyncPaginator

class ModerationProcedures:
    def __init__(self, client: Client):
        self._client = client
        self.legacy: Legacy = Legacy(client)
        self.pull_users: PullUsers = PullUsers(client)
        self.pull: Pull = Pull(client)
        self.note: Note = Note(client)

    def pull_actions(self, sr: str, amount: Optional[int] = None, *,
            action: str = '', mod: str = '',
            ) -> ImpartedPaginatorChainingAsyncIterator[ModerationActionLogAsyncPaginator, ModerationActionLogEntry]:
        p = ModerationActionLogAsyncPaginator(self._client, f'/r/{sr}/about/log', action=action, mod=mod)
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def get_moderator(self, sr: str, user: str) -> Optional[ModeratorUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/moderators', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator_user_item(object_map[order[0]]) if order else None

    async def get_moderator_invitation(self, sr: str, user: str) -> Optional[ModeratorUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/moderators_invited', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator_user_item(object_map[order[0]]) if order else None

    async def get_editable_moderator(self, sr: str, user: str) -> Optional[ModeratorUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/moderators_editable', params={'username': user})
        order = root['moderatorIds']
        object_map = root['moderators']
        return load_moderator_user_item(object_map[order[0]]) if order else None

    async def get_approved_contributor(self, sr: str, user: str) -> Optional[ContributorUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/contributors', params={'username': user})
        order = root['approvedSubmitterIds']
        object_map = root['approvedSubmitters']
        return load_contributor_user_item(object_map[order[0]]) if order else None

    async def get_banned_user(self, sr: str, user: str) -> Optional[BannedUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/banned', params={'username': user})
        order = root['bannedUserIds']
        object_map = root['bannedUsers']
        return load_banned_user_item(object_map[order[0]]) if order else None

    async def get_muted_user(self, sr: str, user: str) -> Optional[MutedUserItem]:
        root = await self._client.request('GET', f'/api/v1/{sr}/muted', params={'username': user})
        order = root['mutedUserIds']
        object_map = root['mutedUsers']
        return load_muted_user_item(object_map[order[0]]) if order else None

    async def send_moderator_invite(self, sr: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'r': sr,
            'type': 'moderator_invite',
            'name': user,
            'permissions': ','.join('+' + p for p in permissions),
        }
        await self._client.request('POST', '/api/friend', data=data)

    async def accept_moderator_invite(self, sr: str, user: str) -> None:
        await self._client.request('POST', '/api/accept_moderator_invite', data={'r': sr})

    async def revoke_moderator_invite(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'moderator_invite',
            'name': user,
        }
        await self._client.request('POST', '/api/friend', data=data)

    async def leave_moderator(self, subreddit_id: int) -> None:
        await self._client.request('POST', '/api/leavemoderator', data={'id': 't5_' + to_base36(subreddit_id)})

    async def remove_moderator(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'moderator',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def set_moderator_permissions(self, sr: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'r': sr,
            'type': 'moderator',
            'name': user,
            'permissions': ','.join('+' + i for i in permissions),
        }
        await self._client.request('POST', '/api/setpermissions', data=data)

    async def set_moderator_invite_permissions(self, sr: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'r': sr,
            'type': 'moderator_invite',
            'name': user,
            'permissions': ','.join('+' + i for i in permissions),
        }
        await self._client.request('POST', '/api/setpermissions', data=data)

    async def add_approved_contributor(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'contributor',
            'name': user,
        }
        await self._client.request('POST', '/api/friend', data=data)

    async def leave_approved_contributor(self, subreddit_id: int) -> None:
        await self._client.request('POST', '/api/leavecontributor', data={'id': 't5_' + to_base36(subreddit_id)})

    async def remove_approved_contributor(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'contributor',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def ban_user(self, sr: str, user: str, *,
            reason: str = '',
            note: str = '',
            duration: Optional[int] = None,
            message: str = '') -> None:
        data = {
            'r': sr,
            'type': 'banned',
            'name': user,
            'ban_reason': reason,
            'note': note,
            'ban_message': message,
        }
        if duration is not None:
            data['duration'] = str(duration)

        await self._client.request('POST', '/api/friend', data=data)

    async def unban_user(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'banned',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def mute_user(self, sr: str, user: str, *, note: str = '') -> None:
        data = {
            'r': sr,
            'type': 'muted',
            'name': user,
            'note': note,
        }
        await self._client.request('POST', '/api/friend', data=data)

    async def unmute_user(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'muted',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def add_approved_wiki_contributor(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'wikicontributor',
            'name': user,
        }
        await self._client.request('POST', '/api/friend', data=data)

    async def remove_approved_wiki_contributor(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'wikicontributor',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    async def ban_wiki_contributor(self, sr: str, user: str, *,
            reason: str = '',
            note: str = '',
            duration: Optional[int] = None) -> None:
        data = {
            'r': sr,
            'type': 'banned',
            'name': user,
            'ban_reason': reason,
            'note': note,
        }
        if duration is not None:
            data['duration'] = str(duration)

        await self._client.request('POST', '/api/friend', data=data)

    async def unban_wiki_contributor(self, sr: str, user: str) -> None:
        data = {
            'r': sr,
            'type': 'wikibanned',
            'name': user,
        }
        await self._client.request('POST', '/api/unfriend', data=data)

    class _removal_reason:
        def __init__(self, outer: ModerationProcedures) -> None:
            self._outer = outer
            self._client = outer._client

        async def create(self, sr: str, title: str, message: str) -> int:
            data = {'title': title, 'message': message}
            root = await self._client.request('POST', f'/api/v1/{sr}/removal_reasons', data=data)
            return int(root['id'], 36)

        async def retrieve_map(self, sr: str) -> Mapping[str, tuple[str, str]]:
            root = await self._client.request('GET', f'/api/v1/{sr}/removal_reasons')
            order = root['order']
            object_map = root['data']
            return {
                y: (m['title'], m['message'])
                for y in order for m in [object_map[y]]
            }

        async def update(self, sr: str, reason_id: str, title: str, message: str) -> None:
            data = {'title': title, 'message': message}
            await self._client.request('PUT', f'/api/v1/{sr}/removal_reasons/{reason_id}', data=data)

        async def delete(self, sr: str, reason_id: str) -> None:
            await self._client.request('DELETE', f'/api/v1/{sr}/removal_reasons/{reason_id}')

    removal_reason: cached_property[_removal_reason] = cached_property(_removal_reason)




from __future__ import annotations
from typing import TypeVar, AsyncIterable, Iterable, AsyncIterator, Iterator

E = TypeVar('E')

class UnfalteringChainingAsyncIterator(AsyncIterator[E]):
    def __init__(self, source: AsyncIterable[Iterable[E]]) -> None:
        self._iterator = source.__aiter__()
        self.current_iter: Iterator[E] = iter(())

    def __aiter__(self) -> AsyncIterator[E]:
        return self

    async def __anext__(self) -> E:
        while True:
            for elem in self.current_iter:
                return elem
            it = await self._iterator.__anext__()
            try:
                self.current_iter = iter(it)
            except StopIteration:
                raise StopAsyncIteration




from typing import Any

# Although the authorisation system used by Reddit's internal API is not
# OAuth2-spec compliant, there are similarities in the way credentials are used.

REDDIT_MOBILE_ANDROID_CLIENT_ID: str = 'ohXpoqrZYub1kg'
REDDIT_MOBILE_IOS_CLIENT_ID: str = 'LNDo9k1o8UAEUw'

ACCOUNTS_BASE_URL: str = "https://accounts.reddit.com"
TOKEN_OBTAINMENT_ENDPOINT_PATH: str = "/api/access_token"
TOKEN_OBTAINMENT_URL: str = ACCOUNTS_BASE_URL + TOKEN_OBTAINMENT_ENDPOINT_PATH

GRANT_DATA: Any = {"scopes":["*","email","pii"]}


GRAPHQL_BASE_URL: str = "https://gql.reddit.com"


SENDBIRD_BASE_URL: str = "https://s.reddit.com"
SENDBIRD_APP_ID: str = "2515BDA8-9D3A-47CF-9325-330BC37ADA13"




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping
if TYPE_CHECKING:
    from ..http.requestor_ASYNC import Requestor
    from ..http.request import Request
    from ..http.response import Response

from urllib.parse import urlsplit

from ..http.requestor_augmenter_ASYNC import RequestorAugmenter

class RedditGiveMeJSONPlease(RequestorAugmenter):
    PARAMS: Mapping[str, str] = {
        'raw_json': '1',
        'api_type': 'json',
    }
    PUBLIC_API_HOST: str = 'oauth.reddit.com'

    def __init__(self, requestor: Requestor) -> None:
        super().__init__(requestor)
        self.params: Mapping[str, str] = dict(self.PARAMS)
        self.public_api_host: str = self.PUBLIC_API_HOST

    async def send(self, request: Request, *, timeout: float = -2) -> Response:
        targeting_public_api = urlsplit(request.uri).netloc == self.public_api_host
        if targeting_public_api:
            rp = request.params
            for k, v in self.params.items():
                if rp.setdefault(k, v) == '\0':
                    del rp[k]

        return await self.requestor.send(request, timeout=timeout)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping, Dict, List
if TYPE_CHECKING:
    from ..client_SYNC import Client
    from ..models.more_comments_SYNC import MoreComments

from ..models.comment_tree_SYNC import MoreCommentsTreeNode, CommentTreeNode, SubmissionTreeNode
from ..models.more_comments_SYNC import ContinueThisThread, LoadMoreComments
from .comment_SYNC import load_comment
from .submission_SYNC import load_submission

def load_more_comments(
    d: Mapping[str, Any],
    client: Client,
    submission_id36: str,
    sort: str,
) -> MoreComments:
    _, _, comment_id36 = d['parent_id'].partition('_')
    go_deeper = d['id'] == '_'
    if go_deeper:
        return ContinueThisThread(
            submission_id36=submission_id36,
            comment_id36=comment_id36,
            sort=sort,
            d=d,
            client=client,
        )
    return LoadMoreComments(
        submission_id36=submission_id36,
        comment_id36=comment_id36,
        child_id36s=d['children'],
        sort=sort,
        count=d['count'],
        d=d,
        client=client,
    )


def load_submission_tree_node(d: Any, client: Client, sort: str) -> SubmissionTreeNode:
    def f(d: Any, client: Client, submission_id36: str, sort: str) -> CommentTreeNode:
        value = load_comment(d, client)
        nodes = []
        more = None

        replies_data = d['replies']
        if replies_data:
            children_data = replies_data['data']['children'].copy()
            if children_data:
                m = children_data[-1]
                if m['kind'] == 'more':
                    children_data.pop()
                    more = load_more_comments(
                        m['data'],
                        client,
                        submission_id36,
                        sort,
                    )

            for m in children_data:
                comment_data = m['data']
                node = f(comment_data, client, submission_id36, sort)
                nodes.append(node)

        return CommentTreeNode(value, nodes, more)

    submission_data = d[0]['data']['children'][0]['data']
    submission_id36: str = submission_data['id']

    value = load_submission(submission_data, client)
    nodes = []
    more = None

    children_data = d[1]['data']['children'].copy()
    if children_data:
        m = children_data[-1]
        if m['kind'] == 'more':
            children_data.pop()
            more = load_more_comments(
                m['data'],
                client,
                submission_id36,
                sort,
            )

    for m in children_data:
        comment_data = m['data']
        node = f(comment_data, client, submission_id36, sort)
        nodes.append(node)

    return SubmissionTreeNode(value, nodes, more)

def load_more_children(d: Any, client: Client, submission_id36: str, sort: str) -> MoreCommentsTreeNode:
    node_lookup: Dict[str, CommentTreeNode] = {}
    chilren_lookup: Dict[str, List[CommentTreeNode]] = {}

    elements = d['json']['data']['things']
    for m in elements:
        kind = m['kind']
        if kind == 'more':
            continue
        data = m['data']
        id_ = data['name']
        value = load_comment(data, client)
        children: List[CommentTreeNode] = []
        node = CommentTreeNode(value, children, None)
        node_lookup[id_] = node
        chilren_lookup[id_] = children

    roots: List[CommentTreeNode] = []
    root_more = None
    for m in elements:
        kind = m['kind']
        data = m['data']
        parent_id = data['parent_id']
        if kind == 'more':
            more = load_more_comments(data, client, submission_id36, sort)
            try:
                node_lookup[parent_id].more = more
            except KeyError:
                if root_more is not None:
                    raise Exception
                root_more = more
        else:
            id_ = data['name']
            node = node_lookup[id_]
            chilren_lookup.get(parent_id, roots).append(node)

    # Use `None` for the node's value here because it's important we
    # don't allow any node to appear more than once in the tree.
    # I.e., we discard the root comment object.
    return MoreCommentsTreeNode(None, roots, root_more)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redditwarp',
 'redditwarp._cli',
 'redditwarp.auth',
 'redditwarp.core',
 'redditwarp.dto_models',
 'redditwarp.dto_models.widget',
 'redditwarp.http',
 'redditwarp.http.misc',
 'redditwarp.http.transport',
 'redditwarp.http.transport.carriers',
 'redditwarp.http.transporters',
 'redditwarp.http.util',
 'redditwarp.internal_api',
 'redditwarp.internal_api.auth',
 'redditwarp.internal_api.core',
 'redditwarp.internal_api.siteprocs',
 'redditwarp.iterators',
 'redditwarp.model_loaders',
 'redditwarp.models',
 'redditwarp.models.widget',
 'redditwarp.pagination',
 'redditwarp.pagination.paginators',
 'redditwarp.pagination.paginators.listing',
 'redditwarp.pagination.paginators.listing.mixins',
 'redditwarp.pagination.paginators.moderation',
 'redditwarp.pagination.paginators.user',
 'redditwarp.pushshift',
 'redditwarp.pushshift.core',
 'redditwarp.pushshift.models',
 'redditwarp.pushshift.paginators',
 'redditwarp.pushshift.siteprocs',
 'redditwarp.siteprocs',
 'redditwarp.siteprocs.account',
 'redditwarp.siteprocs.collection',
 'redditwarp.siteprocs.comment',
 'redditwarp.siteprocs.comment_tree',
 'redditwarp.siteprocs.custom_feed',
 'redditwarp.siteprocs.draft',
 'redditwarp.siteprocs.flair',
 'redditwarp.siteprocs.flair_emoji',
 'redditwarp.siteprocs.front',
 'redditwarp.siteprocs.legacy_reddit_subreddit_style',
 'redditwarp.siteprocs.live_thread',
 'redditwarp.siteprocs.message',
 'redditwarp.siteprocs.misc',
 'redditwarp.siteprocs.moderation',
 'redditwarp.siteprocs.modmail',
 'redditwarp.siteprocs.redesign_reddit_subreddit_style',
 'redditwarp.siteprocs.submission',
 'redditwarp.siteprocs.subreddit',
 'redditwarp.siteprocs.user',
 'redditwarp.siteprocs.widget',
 'redditwarp.siteprocs.wiki',
 'redditwarp.streaming',
 'redditwarp.streaming.carriers',
 'redditwarp.streaming.carriers.main_async',
 'redditwarp.streaming.carriers.main_async.makers',
 'redditwarp.streaming.carriers.main_sync',
 'redditwarp.streaming.carriers.main_sync.makers',
 'redditwarp.streaming.makers',
 'redditwarp.streaming.util',
 'redditwarp.util',
 'redditwarp.websocket',
 'redditwarp.websocket.transport',
 'redditwarp.websocket.transport.carriers',
 'redditwarp.websocket.transporters']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'redditwarp',
    'version': '0.3.0',
    'description': 'The unofficial, type-complete Reddit API library for Python',
    'long_description': None,
    'author': 'Pyprohly',
    'author_email': 'pyprohly@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_collection_ASYNC import SubmissionCollection

from ...model_loaders.submission_collection_ASYNC import load_submission_collection
from ...util.base_conversion import to_base36

class Create:
    def __init__(self, client: Client):
        self._client = client

    async def __call__(self,
        sr_id: int,
        title: str,
        description: Optional[str] = None,
        display_layout: Optional[str] = None,
    ) -> SubmissionCollection:
        sr_id36 = to_base36(sr_id)
        return await self.by_id36(sr_id36, title, description, display_layout)

    async def by_id36(self,
        sr_id36: str,
        title: str,
        description: Optional[str] = None,
        display_layout: Optional[str] = None,
    ) -> SubmissionCollection:
        params = {'sr_fullname': 't5_' + sr_id36, 'title': title}
        if description is not None:
            params['description'] = description
        if display_layout is not None:
            params['display_layout'] = display_layout
        data = await self._client.request('POST', '/api/v1/collections/create_collection', params=params)
        return load_submission_collection(data, self._client)




from __future__ import annotations
from typing import Mapping, Any, Sequence, TypeVar, Generic

from dataclasses import dataclass
from functools import cached_property
from datetime import datetime, timezone

from .artifact import IArtifact, Artifact


# WikiPageRevisionAuthorUser is pretty much the same as User except the
# `awardee_karma`, `awarder_karma`, `total_karma` fields are missing.

class BaseWikiPageRevisionAuthorUser(Artifact):
    class Subreddit:
        def __init__(self, d: Mapping[str, Any]):
            self.name: str = d['display_name']
            self.openness: str = d['subreddit_type']
            self.subscriber_count: int = d['subscribers']
            self.title: str = d['title']
            self.public_description: str = d['public_description']
            self.nsfw: bool = d['over_18']

    def __init__(self, d: Mapping[str, Any]):
        super().__init__(d)
        self.id36: str = d['id']
        self.id: int = int(self.id36, 36)
        self.created_ut: int = int(d['created_utc'])
        self.created_at: datetime = datetime.fromtimestamp(self.created_ut, timezone.utc)
        self.name: str = d['name']

        self.submission_karma: int = d['link_karma']
        self.comment_karma: int = d['comment_karma']

        self.has_premium: bool = d['is_gold']

        self.is_admin: bool = d['is_employee']
        self.is_friend: bool = d['is_friend']
        self.is_a_subreddit_moderator: bool = d['is_mod']

        self.icon_img: str = d['icon_img']

        self.subreddit: BaseWikiPageRevisionAuthorUser.Subreddit = self.Subreddit(d['subreddit'])



TWikiPageRevisionAuthorUser = TypeVar('TWikiPageRevisionAuthorUser', bound=BaseWikiPageRevisionAuthorUser)

@dataclass(repr=False, eq=False)
class GBaseWikiPage(IArtifact, Generic[TWikiPageRevisionAuthorUser]):
    d: Mapping[str, Any]
    body: str
    body_html: str
    can_revise: bool
    revision_uuid: str
    revision_timestamp: int
    revision_author: TWikiPageRevisionAuthorUser
    revision_message: str

    @cached_property
    def revision_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.revision_timestamp, timezone.utc)

@dataclass(repr=False, eq=False)
class BaseWikiPage(GBaseWikiPage[BaseWikiPageRevisionAuthorUser]):
    pass


@dataclass(repr=False, eq=False)
class GBaseWikiPageRevision(IArtifact, Generic[TWikiPageRevisionAuthorUser]):
    d: Mapping[str, Any]
    uuid: str
    timestamp: int
    author: TWikiPageRevisionAuthorUser
    message: str
    hidden: bool

@dataclass(repr=False, eq=False)
class BaseWikiPageRevision(GBaseWikiPageRevision[BaseWikiPageRevisionAuthorUser]):
    pass


@dataclass(repr=False, eq=False)
class GBaseWikiPageSettings(Generic[TWikiPageRevisionAuthorUser]):
    permlevel: int
    editors: Sequence[TWikiPageRevisionAuthorUser]
    unlisted: bool

@dataclass(repr=False, eq=False)
class BaseWikiPageSettings(GBaseWikiPageSettings[BaseWikiPageRevisionAuthorUser]):
    pass




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence
if TYPE_CHECKING:
    from .button import Button

from dataclasses import dataclass

from .widget_SYNC import Widget

@dataclass(repr=False, eq=False)
class ButtonWidget(Widget):
    description: str
    description_html: str
    buttons: Sequence[Button]




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, Mapping
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.live_thread_ASYNC import LiveThread, LiveUpdate

from ...models.live_thread import ContributorList, Contributor
from ...model_loaders.live_thread_ASYNC import load_live_thread, load_live_update
from ...iterators.chunking import chunked
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...util.base_conversion import to_base36
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.live_thread_async1 import LiveUpdateListingAsyncPaginator
from ... import http

class LiveThreadProcedures:
    def __init__(self, client: Client):
        self._client = client

    async def get(self, idt: str) -> Optional[LiveThread]:
        try:
            root = await self._client.request('GET', f'/live/{idt}/about')
        except http.exceptions.StatusCodeException as e:
            if e.status_code == 404:
                return None
            raise
        return load_live_thread(root['data'], self._client)

    def bulk_fetch(self, idts: Iterable[str]) -> CallChunkChainingAsyncIterator[LiveThread]:
        async def mass_fetch(idts: Sequence[str]) -> Sequence[LiveThread]:
            idts_str = ','.join(idts)
            root = await self._client.request('GET', '/api/live/by_id/' + idts_str)
            return [load_live_thread(o['data'], self._client) for o in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, idfs) for idfs in chunked(idts, 100))

    async def create(self, title: str, description: str = '', resources: str = '', *, nsfw: bool = False) -> str:
        form_data = {
            'title': title,
            'description': description,
            'resources': resources,
        }
        if nsfw:
            form_data['nsfw'] = '1'

        root = await self._client.request('POST', '/api/live/create', data=form_data)
        return root['json']['data']['id']

    async def configure(self, idt: str, title: str, description: str, resources: str, nsfw: bool) -> None:
        form_data = {
            'title': title,
            'description': description,
            'resources': resources,
            'nsfw': '01'[nsfw],
        }
        await self._client.request('POST', f'/api/live/{idt}/edit', data=form_data)

    async def close(self, idt: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/close_thread')

    async def get_live_update(self, idt: str, update_uuid: str) -> LiveUpdate:
        root = await self._client.request('GET', f'/live/{idt}/updates/{update_uuid}')
        return load_live_update(root['data'], self._client)

    def pull(self, idt: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingAsyncIterator[LiveUpdateListingAsyncPaginator, LiveUpdate]:
        p = LiveUpdateListingAsyncPaginator(self._client, f'/live/{idt}')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    async def create_live_update(self, idt: str, body: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/update', data={'body': body})

    async def strike_live_update(self, idt: str, update_uuid: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/strike_update', data={'id': 'LiveUpdate_' + update_uuid})

    async def delete_live_update(self, idt: str, update_uuid: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/delete_update', data={'id': 'LiveUpdate_' + update_uuid})

    async def list_contributors(self, idt: str) -> ContributorList:
        root = await self._client.request('GET', f'/live/{idt}/contributors')
        if isinstance(root, Mapping):
            return ContributorList([Contributor(d) for d in root['data']['children']], ())

        contributors = [Contributor(d) for d in root[0]['data']['children']]
        invitations = [Contributor(d) for d in root[0]['data']['children']]
        return ContributorList(contributors, invitations)

    async def send_contributor_invite(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'type': 'liveupdate_contributor_invite',
            'name': user,
            'permissions': ','.join('+' + i for i in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/invite_contributor', data=data)

    async def accept_contributor_invite(self, idt: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/accept_contributor_invite')

    async def revoke_contributor_invite(self, idt: str, user_id36: int) -> None:
        id36 = to_base36(user_id36)
        await self._client.request('POST', f'/api/live/{idt}/rm_contributor_invite', data={'id': 't2_' + id36})

    async def leave_contributor(self, idt: str) -> None:
        await self._client.request('POST', f'/api/live/{idt}/leave_contributor')

    async def remove_contributor(self, idt: str, user_id36: int) -> None:
        id36 = to_base36(user_id36)
        await self._client.request('POST', f'/api/live/{idt}/rm_contributor', data={'id': 't2_' + id36})

    async def set_contributor_permissions(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'type': 'liveupdate_contributor',
            'name': user,
            'permissions': ','.join('+' + i for i in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/set_contributor_permissions', data=data)

    async def set_contributor_invite_permissions(self, idt: str, user: str, permissions: Iterable[str]) -> None:
        data = {
            'type': 'liveupdate_contributor_invite',
            'name': user,
            'permissions': ','.join('+' + i for i in permissions),
        }
        await self._client.request('POST', f'/api/live/{idt}/set_contributor_permissions', data=data)




from __future__ import annotations
from typing import TYPE_CHECKING, Sequence, Optional, Mapping, Any
if TYPE_CHECKING:
    from ..client_ASYNC import Client

from ..exceptions import RejectedResultException
from .comment_tree_ASYNC import MoreCommentsTreeNode

class MoreComments:
    def __init__(self,
        *,
        submission_id36: str,
        comment_id36: str,
        sort: str,
        d: Mapping[str, Any],
        client: Client,
    ):
        self.submission_id36: str = submission_id36
        self.comment_id36: str = comment_id36
        self.sort: str = sort
        self.d: Mapping[str, Any] = d
        self.client: Client = client

    async def __call__(self, *,
        depth: Optional[int] = None,
    ) -> MoreCommentsTreeNode:
        raise NotImplementedError

class ContinueThisThread(MoreComments):
    async def __call__(self, *,
        depth: Optional[int] = None,
    ) -> MoreCommentsTreeNode:
        node = await self.client.p.comment_tree.fetch.by_id36(self.submission_id36, self.comment_id36)
        first_child = node.children[0]
        if first_child.value.id36 != self.comment_id36:
            raise RejectedResultException(f'comment `{self.submission_id36}/{self.comment_id36}` was not found')
        if len(node.children) != 1:
            raise Exception
        return MoreCommentsTreeNode(None, first_child.children, node.more)

class LoadMoreComments(MoreComments):
    def __init__(self,
        *,
        submission_id36: str,
        comment_id36: str,
        child_id36s: Sequence[str],
        sort: str,
        count: int,
        d: Mapping[str, Any],
        client: Client,
    ):
        super().__init__(
            submission_id36=submission_id36,
            comment_id36=comment_id36,
            sort=sort,
            d=d,
            client=client,
        )
        self.child_id36s: Sequence[str] = child_id36s
        self.count: int = count

    async def __call__(self, *,
        depth: Optional[int] = None,
        exact: bool = False,
    ) -> MoreCommentsTreeNode:
        return await self.client.p.comment_tree.more_children.by_id36(
            self.submission_id36,
            self.child_id36s,
            sort=self.sort,
            depth=depth,
            exact=exact,
        )




from __future__ import annotations
from typing import Mapping, Any

from ..models.modmail import ModmailModeratedSubreddit

def load_modmail_moderated_subreddit(d: Mapping[str, Any]) -> ModmailModeratedSubreddit:
    return ModmailModeratedSubreddit(
        d=d,
        id=int(d['id'][3:], 36),
        name=d['name'],
        subscriber_count=d['subscribers'],
    )




from __future__ import annotations
from typing import Any, Mapping
from functools import cached_property

from ..util.data_members_namespace import DataMembersNamespaceMapping
from ..util.attribute_mapping_proxy import DictAndListRecursiveAttributeMappingProxy

class Artifact:
    def __init__(self, d: Mapping[str, Any]):
        self.d: Mapping[str, Any] = d
        self.a: DataMembersNamespaceMapping[object] = DataMembersNamespaceMapping(self)
        self.b: DictAndListRecursiveAttributeMappingProxy[Any] = DictAndListRecursiveAttributeMappingProxy(d)

class IArtifact:
    d: Mapping[str, Any]

    @cached_property
    def a(self) -> DataMembersNamespaceMapping[object]:
        return DataMembersNamespaceMapping(self)

    @cached_property
    def b(self) -> DictAndListRecursiveAttributeMappingProxy[Any]:
        return DictAndListRecursiveAttributeMappingProxy(self.d)




from __future__ import annotations
from typing import Optional, Any, Mapping, Union, \
        IO, Sequence, cast, MutableMapping

import mimetypes
import os.path as op
from itertools import chain

RequestFiles = Union[
    Mapping[str, IO[bytes]],
    Mapping[str, tuple[IO[bytes]]],
    Mapping[str, tuple[IO[bytes], str]],
    Mapping[str, tuple[IO[bytes], str, str]],
    Sequence[tuple[str, IO[bytes], str, str]],
]


class Payload:
    pass


class Content(Payload):
    CONTENT_TYPE_HINT: str = ''

    def get_content_type(self) -> str:
        return self.CONTENT_TYPE_HINT

    def apply_content_type(self, headers: MutableMapping[str, str], *, field_name: str = 'Content-Type') -> None:
        headers.setdefault(field_name, self.get_content_type())

class Bytes(Content):
    CONTENT_TYPE_HINT: str = 'application/octet-stream'

    def __init__(self, data: bytes):
        self.data: bytes = data

class Text(Content):
    CONTENT_TYPE_HINT: str = 'text/plain'

    def __init__(self, text: str):
        self.text: str = text

class URLEncodedFormData(Content):
    CONTENT_TYPE_HINT: str = 'application/x-www-form-urlencoded'

    def __init__(self, data: Mapping[str, str]):
        self.data: Mapping[str, str] = data

class JSON(Content):
    CONTENT_TYPE_HINT: str = 'application/json'

    def __init__(self, json: Any):
        self.json: Any = json



class MultipartFormDataField:
    def __init__(self, name: str):
        self.name: str = name

class MultipartTextField(MultipartFormDataField):
    def __init__(self, name: str, value: str):
        super().__init__(name)
        self.value: str = value

class MultipartFileField(MultipartFormDataField):
    def __init__(self,
            name: str,
            file: IO[bytes],
            filename: str,
            content_type: str):
        super().__init__(name)
        self.file: IO[bytes] = file
        self.filename: str = filename
        self.content_type: str = content_type


class Multipart(Payload):
    CONTENT_TYPE_HINT: str = 'multipart/*'

class MultipartFormData(Multipart):
    CONTENT_TYPE_HINT: str = 'multipart/form-data'

    def __init__(self, parts: Sequence[MultipartFormDataField]):
        self.parts: Sequence[MultipartFormDataField] = parts



def guess_mimetype_from_filename(filename: str) -> str:
    y = mimetypes.guess_type(filename, strict=False)[0]
    return y or 'application/octet-stream'

def make_multipart_form_data_payload(
    data: Optional[Mapping[str, str]],
    files: RequestFiles,
) -> MultipartFormData:
    file_parts: list[MultipartFileField] = []
    if isinstance(files, Mapping):
        for key, value in files.items():
            if isinstance(value, tuple):
                length = len(value)
                if length == 1:
                    file, = cast(tuple[IO[bytes]], value)
                    filename = key
                    content_type = guess_mimetype_from_filename(filename)
                elif length == 2:
                    file, filename = cast(tuple[IO[bytes], str], value)
                    content_type = guess_mimetype_from_filename(filename)
                else:
                    file, filename, content_type = cast(tuple[IO[bytes], str, str], value)

            else:
                file = value
                filename = op.basename(file.name)
                content_type = guess_mimetype_from_filename(filename)

            file_parts.append(MultipartFileField(key, file, filename, content_type))

    else:
        file_parts.extend(
            MultipartFileField(name, file, filename, content_type)
            for (name, file, filename, content_type) in files
        )

    text_parts: list[MultipartTextField] = []
    if data is not None:
        text_parts.extend(MultipartTextField(k, v) for k, v in data.items())

    parts: list[MultipartFormDataField] = list(chain(text_parts, file_parts))
    return MultipartFormData(parts)

def make_payload(
    data: Optional[Union[Mapping[str, str], bytes]] = None,
    json: Any = None,
    files: Optional[RequestFiles] = None,
) -> Optional[Payload]:
    if files is not None:
        if json is not None:
            raise TypeError("`json` and `files` are mutually exclusive parameters")
        if isinstance(data, bytes):
            raise TypeError("`data` cannot be bytes when `files` is used")
        return make_multipart_form_data_payload(data, files)

    if data is not None:
        if json is not None:
            raise TypeError("`json` and `data` are mutually exclusive parameters")
        if isinstance(data, Mapping):
            return URLEncodedFormData(data)
        if isinstance(data, bytes):
            return Bytes(data)
        raise Exception

    if json is not None:
        return JSON(json)

    return None
